from __future__ import print_function

import os
import numpy as np

from petram.model import Model
from .solver_model import Solver

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints("TimeDomainSolver")
rprint = debug.regular_print('TimeDependentSolver')


from petram.solver.std_solver_model import StdSolver
class DerivedValue(StdSolver):
    def allocate_instance(self, engine):
        from petram.solver.std_solver_model import StandardSolver        
        instance = StandardSolver(self, engine)
        instance.set_blk_mask()        
        return instance
    
class TimeStep():
    def __init__(self, data):
        self.data = list(np.atleast_1d(data))
        
    def __call__(self, i):
        if i >= len(self.data):
            return self.data[-1]
        else:
            return self.data[i]        
    
class TimeDomain(Solver):
    can_delete = True
    has_2nd_panel = False

    def attribute_set(self, v):
        v['st_et_nt'] = [0, 1, 5]
        v['time_step'] = 0.01
        v['time_step_cnk'] = 0.01
        v['time_step_fe'] = 0.01                
        v['ts_method'] = "Backward Euler"
        v['abe_minstep']= 0.01
        v['abe_maxstep']= 1.0
        v['use_dwc_cp']   = False   # check point
        v['dwc_cp_name']   = ''              
        v['dwc_cp_arg']   = ''      
        v['use_dwc_ts']   = False   # every time step
        v['dwc_ts_name']   = ''                      
        v['dwc_ts_arg']   = ''      
        
        super(TimeDomain, self).attribute_set(v)
        return v
    
    def panel1_param(self):
        elp_be =  [["dt", "", 0, {}],]
        elp_abe =  [["min. dt", "", 0, {}],
                    ["max. dt", "", 0, {}],]
        ret_cp = [["dwc",   self.dwc_cp_name,   0, {}],
                  ["args.",   self.dwc_cp_arg,   0, {}],]
        value_cp = [self.dwc_cp_name, self.dwc_cp_arg]
        ret_ts = [["dwc",   self.dwc_ts_name,   0, {}],
                  ["args.",   self.dwc_ts_arg,   0, {}],]
        value_ts = [self.dwc_ts_name, self.dwc_ts_arg]
        
        return [#["Initial value setting",   self.init_setting,  0, {},],
                ["physics model",   self.phys_model,  0, {},],
                ["start/end/#step",  "",  0, {},],
                ["probes",   self.probe,  0, {},],                                
                [None, None, 34, ({'text':'method','choices': ["Backward Euler",
                                                               "CrankNicolson",
                                                               "Forward Euler",
                                                               "Adaptive BE"], 'call_fit':False},
                                  {'elp':elp_be},
                                  {'elp':elp_be},
                                  {'elp':elp_be},                                                                    
                                  {'elp':elp_abe},)],
                [None, [False, value_cp], 27, [{'text':'Use DWC (check point)'},
                                              {'elp': ret_cp}]],
                [None, [False, value_ts], 27, [{'text':'Use DWC (time stepping)'},
                                              {'elp': ret_ts}]],
                [None,
                 self.clear_wdir,  3, {"text":"clear working directory"}],
                [None,
                 self.init_only,  3, {"text":"initialize solution only"}], 
                [None,
                 self.assemble_real,  3, {"text":"convert to real matrix (complex prob.)"}],
                [None,
                 self.save_parmesh,  3, {"text":"save parallel mesh"}],
                [None,
                 self.use_profiler,  3, {"text":"use profiler"}],]

    def get_panel1_value(self):
        st_et_nt = ", ".join([str(x) for x in self.st_et_nt])
        return (#self.init_setting,
                self.phys_model,
                st_et_nt,
                self.probe,                                
                [self.ts_method,
                 [str(self.time_step),],
                 [str(self.time_step_cnk),],
                 [str(self.time_step_fe),],                                  
                 [str(self.abe_minstep), str(self.abe_maxstep),],
                ],
                [self.use_dwc_cp, [self.dwc_cp_name, self.dwc_cp_arg,]],
                [self.use_dwc_ts, [self.dwc_ts_name, self.dwc_ts_arg,]],            
                self.clear_wdir,
                self.init_only,               
                self.assemble_real,
                self.save_parmesh,
                self.use_profiler,)

    
    def import_panel1_value(self, v):
        #self.init_setting = str(v[0])        
        self.phys_model = str(v[0])
        tmp = str(v[1]).split(',')
        st_et_nt = [tmp[0], tmp[1], ",".join(tmp[2:])]
        self.st_et_nt = [eval(x) for x in st_et_nt]
        self.probe = str(v[2])
        self.clear_wdir = v[6]
        self.init_only = v[7]        
        self.assemble_real = v[8]
        self.save_parmesh = v[9]
        self.use_profiler = v[10]
        
        self.ts_method = str(v[3][0])
        self.time_step = str(v[3][1][0])
        self.time_step_cnk = str(v[3][2][0])
        self.time_step_fe = str(v[3][3][0])                
        self.abe_minstep  = float(v[3][4][0])
        self.abe_maxstep  = float(v[3][4][1])
        self.use_dwc_cp   = v[4][0]
        self.dwc_cp_name  = v[4][1][0]                           
        self.dwc_cp_arg   = v[4][1][1]         
        self.use_dwc_ts   = v[5][0]
        self.dwc_ts_name  = v[5][1][0]
        self.dwc_ts_arg   = v[5][1][1]                           
        

    def get_possible_child(self):
        choice = []
        try:
            from petram.solver.mumps_model import MUMPS
            choice.append(MUMPS)
        except ImportError:
            pass

        try:
            from petram.solver.iterative_model import Iterative
            choice.append(Iterative)
        except ImportError:
            pass

        try:
            from petram.solver.strumpack_model import SpSparse
            choice.append(SpSparse)
        except ImportError:
            pass
        choice.append(DerivedValue)
        return choice
    
    def get_matrix_weight(self, timestep_config):
        #timestep_weight):
        #dt = float(self.time_step)
        #lns = self.root()['General']._global_ns.copy()
        #lns['dt'] = dt
        wt = [1 if x else 0 for x in timestep_config]
        return wt

    def get_child_solver(self):
        return self.derived_value_solver()
    
    def derived_value_solver(self):
        return [self[key] for key in self
                if isinstance(self[key],DerivedValue) and self[key].enabled]
    
    @debug.use_profiler
    def run(self, engine, is_first=True):
        if self.clear_wdir:
            engine.remove_solfiles()

        fid = engine.open_file('checkpoint.'+self.parent.name()+'_'+self.name()+'.txt', 'w')
        st, et, nt = self.st_et_nt
        
        if self.ts_method == 'Backward Euler' or self.ts_method == 'Backward Eular':
            instance = FirstOrderBackwardEuler(self, engine)
            time_step = self.eval_text_in_global(self.time_step)
            dprint1("time step configuration: " + str(self.time_step) + ':' +  str(time_step))            
            instance.set_timestep(TimeStep(time_step))
            
        elif self.ts_method == "CrankNicolson":
            instance = CrankNicolson(self, engine)
            time_step = self.eval_text_in_global(self.time_step_cnk)
            dprint1("time step configuration: " + str(self.time_step_cnk) + ':' +  str(time_step))            
            instance.set_timestep(TimeStep(time_step))
            
        elif self.ts_method == "Forward Euler":
            instance = FirstOrderForwardEuler(self, engine)            
            time_step = self.eval_text_in_global(self.time_step_fe)
            dprint1("time step configuration: " + str(self.time_step_fe) + ':' +  str(time_step))            
            instance.set_timestep(TimeStep(time_step))
            
        elif self.ts_method == "Adaptive BE":
            instance = FirstOrderBackwardEulerAT(self, engine)
            instance.set_timestep(self.abe_minstep)
            instance.set_maxtimestep(self.abe_maxstep)
        else:
            assert False, "unknown stepping method: "+ self.ts_method
            
        instance.set_start(st)
        instance.set_end(et)
        instance.set_checkpoint(np.linspace(st, et, nt))

        engine.sol = engine.assembled_blocks[1][0]
        instance.sol = engine.sol
        instance.time = st
        
        if self.init_only:
            instance.write_checkpoint_solution()
        
        else:
            if is_first:
                instance.pre_assemble()                                
                instance.assemble()
                
            #instance.solve()            
            #if is_first:
            #self.prepare_form_sol_variables(engine)
            #finished = instance.init(self.init_only)
            
            instance.set_blk_mask()                        
            instance.configure_probes(self.probe)
        
            for solver in self.derived_value_solver():
                child = solver.allocate_instance(engine)
                instance.add_child_instance(child)
                
            finished = False
            if fid is not None:
                fid.write(str(0)+':'+str(instance.time)+"\n")
            while not finished:
                finished, cp_written = instance.step(is_first)
                is_first=False
                
                if self.use_dwc_ts:
                    engine.call_dwc(self.get_phys_range(),
                                    method="timestep",
                                    callername = self.name(),
                                    dwcname = self.dwc_ts_name,
                                    args = self.dwc_ts_arg,
                                    time = instance.time)
                if self.use_dwc_cp and cp_written:                
                    engine.call_dwc(self.get_phys_range(),
                                    method="checkpoint",
                                    callername = self.name(),
                                    dwcname = self.dwc_cp_name, 
                                    args = self.dwc_cp_arg,
                                    time = instance.time,
                                    icheckpoint = instance.icheckpoint-1)
                if cp_written:
                    instance.save_probe()
                    if fid is not None:
                        fid.write(str(instance.icheckpoint-1)+':'+str(instance.time)+"\n")
                        fid.flush()
                        
        instance.save_solution(ksol = 0,
                               skip_mesh = False, 
                               mesh_only = False,
                               save_parmesh=self.save_parmesh)
        instance.save_probe()
        if fid is not None: fid.close()
        
        return is_first       

from petram.solver.solver_model import TimeDependentSolverInstance

class FirstOrderBackwardEuler(TimeDependentSolverInstance):
    '''
    Fixed time step solver
    '''
    def __init__(self, gui, engine):
        TimeDependentSolverInstance.__init__(self, gui, engine)
        self.pre_assembled = False
        self.assembled = False
        self.counter = 0
        self._dt_used_in_assemble = 0.0        
    @property
    def time_step(self):
        return self._time_step(self.counter)

    def set_blk_mask(self):
        super(FirstOrderBackwardEuler, self).set_blk_mask()
        phys_target = self.get_target_phys()
        time_deriv_vars = []
        for phys in phys_target:
           dep_vars0 = phys.dep_vars0
           dep_vars = phys.dep_vars
           for x in dep_vars:
               if not x in dep_vars0: time_deriv_vars.append(x)
        self.time_deriv_vars = time_deriv_vars
        
        #self.fes_dt_mask = [False]*len(self.fes_mask)
        #for name in time_deriv_vars:
        #     offset = self.engine.dep_var_offset(name)
        #     self.fes_dt_mask[offset] = True

        dprint1("time deivatives to be computed", time_deriv_vars)
        
    def pre_assemble(self, update=False):
        engine = self.engine
        phys_target = self.get_phys()
        phys_range  = self.get_phys_range()
        
        if not update:
            engine.run_verify_setting(phys_target, self.gui)
            
        M_Updated = engine.run_assemble_mat(phys_target, phys_range,
                                             update=update)
        B_Updated = engine.run_assemble_b(phys_target, update=update)
        self.pre_assembled = True

        return M_Updated, B_Updated


    def compute_A(self, M, B, X, mask_M, mask_B):
        '''
        M/dt u_1 + K u_1 = M/dt u_0 + b
        '''
        one_dt = 1./float(self.time_step)
        A = M[0]+ M[1]*one_dt
        dprint1("A", A)
        return A, np.any(mask_M)

    def compute_rhs(self, M, B, X):
        one_dt = 1./float(self.time_step)
        MM = M[1]*one_dt
        RHS = MM.dot(self.engine.sol) + B
        dprint1("RHS", RHS)
        return RHS

    def assemble(self, update=False):
        blocks, M_changed = self.engine.run_assemble_blocks(self.compute_A,
                                        self.compute_rhs,
                                        inplace=False,
                                        update=update)

        #A, X, RHS, Ae, B, M, depvars = blocks
        self.assembled = True
        self._dt_used_in_assemble = self.time_step
        return M_changed
        
    def step(self, is_first):
        engine = self.engine
        mask = self.blk_mask
        engine.copy_block_mask(mask)
        
        #if not self.pre_assembled:
        #    assert False, "pre_assmeble must have been called"
            
        if (self.counter == 0 and is_first):
            M_changed = True
        else:
            if self._dt_used_in_assemble != self.time_step:
                engine.set_update_flag('UpdateAll')
            else:
                engine.set_update_flag('TimeDependent')    
            engine.run_apply_essential(self.get_phys(), self.get_phys_range(),
                                       update=True)
            engine.run_fill_X_block(update=True)        
            M_updated, B_updated = self.pre_assemble(update=True)
            M_changed = self.assemble(update=True)

        A, X, RHS, Ae, B, M, depvars = self.blocks
        if M_changed or self.counter == 0:
            AA = engine.finalize_matrix(A, mask,
                                        not self.phys_real, format = self.ls_type,
                                        verbose=False)
        BB = engine.finalize_rhs([RHS], A, X[-1], mask,
                                     not self.phys_real, format = self.ls_type,
                                     verbose=False)
        if self.counter == 0:
            self.sol = engine.sol
            self.write_checkpoint_solution()
            self.icheckpoint += 1

        depvars = [x for i, x in enumerate(depvars) if mask[0][i]]
        if self.linearsolver is None:
            is_complex = self.gui.is_complex()
            self.linearsolver  = self.linearsolver_model.allocate_solver(is_complex, engine)
            M_changed = True
            
        if M_changed:
            self.linearsolver.SetOperator(AA, dist = engine.is_matrix_distributed,
                                          name = depvars)
            
        if self.linearsolver.is_iterative:
            XX = engine.finalize_x(X[-1], RHS, mask, not self.phys_real,
                                   format = self.ls_type)
        else:
            XX = None
        solall = self.linearsolver.Mult(BB, x=XX, case_base=engine.case_base)
        engine.case_base += len(BB)
            
        if not self.phys_real and self.assemble_real:
            assert False, "this has to be debugged (convertion from real to complex)"
            solall = self.linearsolver_model.real_to_complex(solell, A)

        A.reformat_central_mat(solall, 0, X[0], mask)

        # this apply interpolation operator 
        sol, sol_extra = engine.split_sol_array(X[0])

        for name in self.time_deriv_vars:
            offset1 = engine.dep_var_offset(name)       # vt
            offset2 = engine.dep_var_offset(name[:-1])  # v
            X[0][offset1, 0] = (X[0][offset2, 0]-X[-1][offset2, 0])*(1./self.time_step)

        for child in self.child_instance:
            # for now update_operator is True only for the first run.
            child.solve(update_operator = (self.counter == 0))

        self.time = self.time + self.time_step
        
        self.counter += 1
        for p in self.probe:
            p.append_sol(X[0], self.time)

        
        # swap X[0] and X[-1] for next computing
        tmp = X[0]; X[0] = X[-1]; X[-1]=tmp
        self.sol = X[-1]
        engine.sol = self.sol

        engine.recover_sol(sol, access_idx=-1)
        ## ToDo. Provide a way to use Lagrange multipler in model
        extra_data = engine.process_extra(sol_extra)

        checkpoint_written = False
        if self.checkpoint[self.icheckpoint] < self.time:
            self.write_checkpoint_solution()
            self.icheckpoint += 1
            checkpoint_written = True
            
        dprint1("TimeStep ("+str(self.counter)+ "), t="+str(self.time)+"...done.")
        dprint1(debug.format_memory_usage())        
        return self.time >= self.et, checkpoint_written

    def write_checkpoint_solution(self):
        dprint1("writing checkpoint t=" + str(self.time) +
                "("+str(self.icheckpoint)+")")        
        od = os.getcwd()
        path = os.path.join(od, 'checkpoint_' + self.gui.parent.name()+'_'+
                            self.gui.name()+'_'+str(self.icheckpoint))
        self.engine.mkdir(path) 
        os.chdir(path)
        self.engine.cleancwd() 
        self.save_solution()
        self.engine.symlink('../model.pmfm', 'model.pmfm')        
        os.chdir(od)
        
class CrankNicolson(FirstOrderBackwardEuler):
    def compute_A(self, M, B, X, mask_M, mask_B):
        '''
        M/dt u_1 + K/2 u_1 = M/dt u_0 - K/2 u_0 + b
        '''
        one_dt = 1./float(self.time_step)
        #MM = M[1]*one_dt
        A = M[0]*0.5+ M[1]*one_dt
        dprint1("A", A)
        return A, np.any(mask_M)

    def compute_rhs(self, M, B, X):
        one_dt = 1./float(self.time_step)
        MM = (-M[0]*0.5 + M[1]*one_dt)
        RHS = MM.dot(self.engine.sol) + B
        dprint1("RHS", RHS)
        return RHS

class FirstOrderForwardEuler(FirstOrderBackwardEuler):
    def compute_A(self, M, B, X, mask_M, mask_B):
        '''
        M/dt u_1  = M/dt u_0 - K u_0 + b
        '''
        one_dt = 1./float(self.time_step)
        #MM = M[1]*one_dt
        A = M[1]*one_dt
        dprint1("A", A)
        return A, np.any(mask_M)

    def compute_rhs(self, M, B, X):
        one_dt = 1./float(self.time_step)
        MM = (-M[0] + M[1]*one_dt)
        RHS = MM.dot(self.engine.sol) + B
        dprint1("RHS", RHS)
        return RHS
    
    
class FirstOrderBackwardEulerAT(FirstOrderBackwardEuler):
    def __init__(self, gui, engine):
        FirstOrderBackwardEuler.__init__(self, gui, engine)
        self.linearsolver  = {}
        self.blocks1 = {}
        self.sol1 = None
        self.sol2 = None
        self._time_step1 = 0
        self._time_step2 = -1
        self.maxstep = 0

    def set_maxtimestep(self, dt):
        self.max_timestep = dt
    @property
    def time_step1(self):
        return (self.time_step_base)* 2**self._time_step1
    @property
    def time_step2(self):
        return (self.time_step_base)* 2**self._time_step2
    
    def set_timestep(self, time_step):
        self.time_step = time_step
        self.time_step_base = time_step
        
    def assemble(self,idt=None):
        flag = False
        if idt is None:
            idt = 0
            flag = True
        self.blocks1[idt] = self.engine.run_assemble_blocks(self.compute_A,
                                                            self.compute_rhs,
                                                            inplace=False)        
        if flag:
            self.blocks = self.blocks1[0]
        else:
            self.blocks = None
            
    def step(self):
        dprint1("Entering step", self.time_step1)
        def get_A_BB(mode, sol, recompute_rhs=False):
            if sol is None: sol = self.sol
            idt = self._time_step1 if mode == 0 else self._time_step2
            dt  = self.time_step1 if mode == 0 else self.time_step2
            self.time_step = dt            
            if not idt in self.blocks1:
                self.assemble(idt = idt)
                A, X, RHS, Ae, B, M, depvars = self.blocks1[idt]
                BB = engine.finalize_rhs([RHS], not self.phys_real,
                                         format = self.ls_type, verbose=False)
            else:
                A, X, RHS, Ae, B, M, depvars = self.blocks1[idt]
                if self.counter != 0 or recompute_rhs:
                    # recompute RHS
                    RHS = self.compute_rhs(M, B, [sol])
                    RHS = engine.eliminateBC(Ae, X[1], RHS)
                    RHS = engine.apply_interp(RHS=RHS)
                BB = engine.finalize_rhs([RHS], not self.phys_real,
                                         format = self.ls_type, verbose=False)
            if not idt in self.linearsolver:
                AA = engine.finalize_matrix(A, not self.phys_real,
                                            format = self.ls_type, verbose=False)
                if self.ls_type.startswith('coo'):
                    datatype = 'Z' if (AA.dtype == 'complex') else 'D'
                else:
                    datatype = 'D'
                self.linearsolver[idt]  = self.linearsolver_model.allocate_solver(datatype,
                                                                                  engine)
                self.linearsolver[idt].SetOperator(AA,
                                                   dist = engine.is_matrix_distributed,
                                                   name = depvars)
                
            return A, BB
            
        engine = self.engine

        if not self.pre_assembled:
            assert False, "pre_assmeble must have been called"

        A, BB = get_A_BB(0, self.sol1)
        solall = self.linearsolver[self._time_step1].Mult(BB)
        sol1 = A.reformat_central_mat(solall, 0)
        print("check sample1 (0)", [p.current_value(sol1) for p in self.probe])
        
        A, BB = get_A_BB(1, self.sol2)
        solall2 = self.linearsolver[self._time_step2].Mult(BB)
        sol2 = A.reformat_central_mat(solall2, 0)
        print("check sample2 (1)", [p.current_value(sol2) for p in self.probe])
        
        A, BB = get_A_BB(1, sol2, recompute_rhs=True)
        solall2 = self.linearsolver[self._time_step2].Mult(BB)
        sol2 = A.reformat_central_mat(solall2, 0)
        print("check sample2 (1)", [p.current_value(sol2) for p in self.probe])

        sample1 = np.hstack([p.current_value(sol1) for p in self.probe]).flatten()
        sample2 = np.hstack([p.current_value(sol2) for p in self.probe]).flatten()
        
        from petram.mfem_config import use_parallel
        if use_parallel:
             from petram.helper.mpi_recipes import allgather, allgather_vector
             sample1 = allgather_vector(np.atleast_1d(sample1))
             sample2 = allgather_vector(np.atleast_1d(sample2))
             
        delta=np.mean(np.abs(sample1-sample2)/np.abs(sample1+sample2)*2)
    
        threshold = 0.01
        if delta > threshold:
            dprint1("delta is large ", delta, sample1, sample2)
            if self._time_step1 > 0:
                self._time_step1 -= 1
                self._time_step2 -= 1
                dprint1("next try....", self.time_step1, self.time_step2)
                return False                
            else:
                dprint1("delta may be too large, but restricted by min time step")                
        else:
            dprint1("delta good  ", delta, sample1, sample2)
        
        if not self.phys_real and self.assemble_real:
            assert False, "this has to be debugged (convertion from real to complex)"
            solall = self.linearsolver_model.real_to_complex(solell, A)
            
        self.time = self.time + self.time_step1
        self.counter += 1
        
        self.sol = A.reformat_central_mat(solall, 0)
        self.sol1 = self.sol
        self.sol2 = self.sol        

        for p in self.probe:
            p.append_sol(self.sol, self.time)
                
        if self.checkpoint[self.icheckpoint] < self.time:
            self.write_checkpoint_solution()
            self.icheckpoint += 1

        dprint1("TimeStep ("+str(self.counter)+ "), t="+str(self.time)+"...done.")
        if delta < threshold/4:
            dt_test = (self.time_step_base)* 2**(self._time_step1 + 1)
            if self.max_timestep >  dt_test:
                 self._time_step1 += 1
                 self._time_step2 += 1
                 dprint1("next try....", self.time_step1, self.time_step2)
            else:
                 dprint1("delta is small, but restricted by max time step")
        return self.time >= self.et
                      






from __future__ import print_function
from petram.helper.matrix_file import write_matrix, write_vector, write_coo_matrix
from petram.mfem_config import use_parallel
from .solver_model import LinearSolverModel, LinearSolver
from .solver_model import Solver
from petram.namespace_mixin import NS_mixin


import sys
import numpy as np
import scipy
from scipy.sparse import coo_matrix, csr_matrix

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('StrumpackModel')


if use_parallel:
    from petram.helper.mpi_recipes import *
    from mfem.common.parcsr_extra import *
    import mfem.par as mfem
    default_kind = 'hypre'

    from mpi4py import MPI
    num_proc = MPI.COMM_WORLD.size
    myid = MPI.COMM_WORLD.rank
    smyid = '{:0>6d}'.format(myid)
    from mfem.common.mpi_debug import nicePrint

else:
    import mfem.ser as mfem
    default_kind = 'scipy'
    num_proc = 1
    myid = 0

    def nicePrint(*x):
        print(x)


attr_names = ['log_level',
              'ordering',
              'mc64job',
              'compression',
              'compression_rel_tol',
              'compression_abs_tol',
              'lossy_precision',
              'krylov',
              'rctol',
              'actol',
              'gmres_restart',
              'gstype',
              'maxiter',
              'use_gpu',
              'cuda_cutoff',
              'cuda_streams',
              'MUMPS_SYMQAMD',
              'agg_amalg',
              'indirect_sampling',
              'replace_tiny_pivots',
              'compression_min_sep_size',
              'compression_min_front_size',
              'compression_leaf_size',
              'separator_ordering_level',
              'hodlr_butterfly_levels',
              'use_single_precision',
              'use_64_int',
              'write_mat',
              'extra_options',
              'use_dist_sol', ]

ordering_modes = {"Natural": "STRUMPACK_NATURAL",
                  "Metis": "STRUMPACK_METIS",
                  "ParMetis": "STRUMPACK_PARMETIS",
                  "Scotch": "STRUMPACK_SCOTCH",
                  "PT-Scotch": "STRUMPACK_PTSCOTCH",
                  "RCM": "STRUMPACK_RCM"}

krylov_modes = {"Auto": "STRUMPACK_AUTO",
                "Direct": "STRUMPACK_DIRECT",
                "Refinement": "STRUMPACK_REFINE",
                "PGMRES": "STRUMPACK_PREC_GMRES",
                "GMRES": "STRUMPACK_GMRES",
                "PBICGSTAB": "STRUMPACK_PREC_BICGSTAB",
                "BICGSTAB": "BICGSTAB", }

compression_modes = {"None": "STRUMPACK_NONE",
                     "HSS": "STRUMPACK_HSS",
                     "BLR": "STRUMPACK_BLR",
                     "HOLDR": "STRUMPACK_HODLR",
                     "HOLBF": "STRUMPACK_HODLR",  # this is special
                     "LOESSLESS": "STRUMPACK_LOSSLESS",
                     "LOSSY": "STRUMPACK_LOSSY", }

GramSchmidt_types = {"Classical": "STRUMPACK_CLASSICAL",
                     "Modified": "STRUMPACK_MODIFIED"}


help_txt = ("0: no reordering for stability, this disables MC64/matching",
            "1: MC64(1): currently not supported",
            "2: MC64(2): maximize the smallest diagonal value",
            "3: MC64(3): maximize the smallest diagonal value, different strategy",
            "4: MC64(4): maximize sum of diagonal values",
            "5: MC64(5): maximize product of diagonal values and apply row and column scaling",
            "6: Combinatorial BLAS: approximate weight perfect matching",)


class Strumpack(LinearSolverModel):
    hide_ns_menu = True
    has_2nd_panel = False
    accept_complex = True
    always_new_panel = False

    def __init__(self, *args, **kwargs):
        LinearSolverModel.__init__(self, *args, **kwargs)

    def init_solver(self):
        pass

    def panel1_param(self):
        return [
            ["log_level", self.log_level, 400, {}],
            ["ordering", self.ordering, 4, {"readonly": True,
                                            "choices": list(ordering_modes), }],
            ["mc64 matching", self.mc64job, 4, {"readonly": True,
                                                "choices": ["0 (default)", "2", "3", "4", "5", "6"], }],
            ["compression", self.compression, 4, {"readonly": True,
                                                  "choices": list(compression_modes)}, ],
            ["compression rel. tol.", self.compression_rel_tol, 0, {}],
            ["compression abs. tol.", self.compression_abs_tol, 0, {}],
            ["lossy precision", self.lossy_precision, 0, {}],
            ["Krylov", self.krylov, 4, {"readonly": True,
                                        "choices": list(krylov_modes)}],
            ["rctol", self.rctol, 0, {}],
            ["actol", self.actol, 0, {}],
            ["gmres restrart", self.gmres_restart, 0, {}],
            ["GramSchmidt type", self.gstype, 4, {"readonly": True,
                                                  "choices": list(GramSchmidt_types), }],
            ["max iter.", self.maxiter, 0, {}],
            ["GPU", self.use_gpu, 3, {"text": ""}],
            ["CUDA cutoff", self.cuda_cutoff, 0, {}],
            ["CUDA streams", self.cuda_streams, 0, {}],
            ["MUMPS_SYMQAMD", self.MUMPS_SYMQAMD, 3, {"text": ""}],
            ["agg_amalg", self.agg_amalg, 3, {"text": ""}],
            ["indirect_sampling", self.indirect_sampling, 3, {"text": ""}],
            ["replace_tiny_pivots", self.replace_tiny_pivots, 3, {"text": ""}],
            ["compression min sep", self.compression_min_sep_size, 0, {}],
            ["compression min front", self.compression_min_front_size, 0, {}],
            ["compression leaf size", self.compression_leaf_size, 0, {}],
            ["separator ordering level", self.separator_ordering_level, 0, {}],
            ["hodlr butterfly levels", self.hodlr_butterfly_levels, 0, {}],
            ["single preceision", self.use_single_precision, 3, {"text": ""}],
            ["use 64bit integer", self.use_64_int, 3, {"text": ""}],
            ["write matrix", self.write_mat, 3, {"text": ""}],
            ["extra options", self.extra_options, 2235, {'nlines': 3}, ],
            ["use dist, SOL (dev.)", self.use_dist_sol, 3, {"text": ""}], ]

    def get_panel1_value(self):
        ans = []
        for n, p in zip(attr_names, self.panel1_param()):
            value = getattr(self, n)
            if p[3] == 3:
                value = bool(value)
            elif p[3] == 400:
                value = int(value)
            ans.append(value)
        return ans

    def import_panel1_value(self, v):
        for value, n, p in zip(v, attr_names, self.panel1_param()):
            if p[3] == 3:
                value = bool(value)
            if p[3] == 400:
                value = int(value)
            setattr(self, n, value)

    def attribute_set(self, v):
        v = super(Strumpack, self).attribute_set(v)
        v['log_level'] = 0
        v['write_mat'] = False
        v['gstype'] = 'Classical'
        v['rctol'] = "1e-6 (default)"
        v['actol'] = "1e-10 (default)"
        v['maxiter'] = "5000 (default)"
        v['gmres_restart'] = "30 (default)"
        v['mc64job'] = "0 (default)"
        v['krylov'] = 'Auto'
        v['use_gpu'] = False
        v['ordering'] = "Metis"
        v['compression'] = "None"
        v['use_single_precision'] = False
        v["cuda_cutoff"] = "500 (default)"
        v["cuda_streams"] = "10 (default)"
        v["MUMPS_SYMQAMD"] = False
        v["agg_amalg"] = False
        v["indirect_sampling"] = False
        v["replace_tiny_pivots"] = False
        v["compression_min_sep_size"] = "2147483647 (default)"
        v["compression_min_front_size"] = "2147483647 (default)"
        v["compression_leaf_size"] = "2147483647 (default)"
        v["separator_ordering_level"] = "1 (default)"
        v["hodlr_butterfly_levels"] = "100 (default)"
        v["compression_rel_tol"] = "1e-4 (default)"
        v["compression_abs_tol"] = "1e-8 (default)"
        v["lossy_precision"] = "16 (default)"
        v["extra_options"] = ""
        v["use_64_int"] = False
        v['use_dist_sol'] = True

        return v

    def does_linearsolver_choose_linearsystem_type(self):
        return True

    def linear_system_type(self, assemble_real, phys_real):
        if phys_real:
            return 'blk_interleave'
        else:
            return 'blk_merged'

    def real_to_complex(self, solall, M):
        if self.parent.assemble_real:
            return self.real_to_complex_merged(solall, M)
        else:
            assert False, "should not come here"

    def real_to_complex_merged(self, solall, M):
        if use_parallel:
            from mpi4py import MPI
            myid = MPI.COMM_WORLD.rank

            offset = M.RowOffsets().ToList()
            of = [np.sum(MPI.COMM_WORLD.allgather(np.int32(o)))
                  for o in offset]
            if myid != 0:
                return

        else:
            offset = M.RowOffsets()
            of = offset.ToList()
        dprint1(of)
        rows = M.NumRowBlocks()
        s = solall.shape
        i = 0
        pt = 0
        result = np.zeros((s[0] // 2, s[1]), dtype='complex')
        for i in range(rows):
            l = of[i + 1] - of[i]
            w = int(l // 2)
            result[pt:pt + w, :] = (solall[of[i]:of[i] + w, :]
                                    + 1j * solall[(of[i] + w):of[i + 1], :])
            pt = pt + w
        return result

    def allocate_solver(self, is_complex=False, engine=None):
        solver = StrumpackSolver(self, engine)
        solver.AllocSolver(is_complex, self.use_single_precision)
        return solver


def get_block(Op, i, j):
    try:
        return Op._linked_op[(i, j)]
    except KeyError:
        return None


def build_csr_local(A, dtype, is_complex):
    '''
    build CSR form of A as a single
    matrix
    '''
    # print("build_csr_local", dtype, is_complex)
    offset = np.array(A.RowOffsets().ToList(), dtype=int)
    coffset = np.array(A.ColOffsets().ToList(), dtype=int)
    if is_complex:
        offset = offset // 2
    # nicePrint("offets", offset, coffset)
    rows = A.NumRowBlocks()
    cols = A.NumColBlocks()

    local_size = np.diff(offset)
    # nicePrint("local_size",local_size)

    if use_parallel:
        x = allgather_vector(local_size)
        global_size = np.sum(x.reshape(num_proc, -1), 0)
        global_offset = np.hstack(([0], np.cumsum(global_size)))
        # global_roffset = global_offset + offset
        # nicePrint("global offset/roffset", global_offset)
        new_offset = np.hstack(([0], np.cumsum(x)))[:-1]
        new_size = x.reshape(num_proc, -1)
        new_offset = new_offset.reshape(num_proc, -1)
        # nicePrint(new_size)
        # nicePrint(new_offset)
    else:
        global_size = local_size
        new_size = local_size.reshape(1, -1)
        new_offset = offset.reshape(1, -1)

    # index_mapping
    def blk_stm_idx_map(i):
        stm_idx = [new_offset[kk, i] +
                   np.arange(new_size[kk, i], dtype=int)
                   for kk in range(len(new_offset))]
        return np.hstack(stm_idx)

    def sparsemat2csr(m):
        w, h = m.Width(), m.Height()
        I = m.GetIArray()
        J = m.GetJArray()
        data = m.GetDataArray()
        m = csr_matrix((data, J, I), shape=(h, w),
                       dtype=data.dtype)
        return m

    def ToScipyCoo(mat):
        '''
        convert HypreParCSR to Scipy Coo Matrix
        '''
        num_rows, ilower, iupper, jlower, jupper, irn, jcn, data = mat.GetCooDataArray()
        m = iupper - ilower + 1
        n = mat.N()

        return coo_matrix((data, (irn - ilower, jcn)),
                          shape=(num_rows, n)), ilower

    map = [blk_stm_idx_map(i) for i in range(rows)]
    # nicePrint("map", map)
    newi = []
    newj = []
    newd = []
    nrows = np.sum(local_size)
    ncols = np.sum(global_size)

    from scipy.sparse import bmat

    elements = [None] * rows
    elements = [elements.copy() for x in range(cols)]
    for i in range(rows):
        for j in range(cols):
            m = get_block(A, i, j)
            if m is None:
                continue
            if use_parallel:
                if isinstance(m, mfem.ComplexOperator):
                    if not is_complex:
                        mr, ilower = ToScipyCoo(m._real_operator)
                        mi, ilower = ToScipyCoo(m._imag_operator)
                        m = bmat([[mr, -mi], [mi, mr]])
                    else:
                        mr, ilower = ToScipyCoo(m._real_operator)
                        mi, ilower = ToScipyCoo(m._imag_operator)
                        m = (mr + 1j * mi).tocoo()
                else:
                    m, ilower = ToScipyCoo(m)
            else:
                # this is not efficient but for now let's do this...
                if isinstance(m, mfem.ComplexOperator):
                    if not is_complex:
                        mr = m._real_operator
                        mi = m._imag_operator
                        mr = sparsemat2csr(mr)
                        mi = sparsemat2csr(mi)
                        m = bmat([[mr, -mi], [mi, mr]]).tocoo()
                    else:
                        mr = m._real_operator
                        mi = m._imag_operator
                        mr = sparsemat2csr(mr)
                        mi = sparsemat2csr(mi)
                        m = (mr + 1j * mi).tocoo()
                else:
                    m = sparsemat2csr(m).tocoo()

            elements[i][j] = m
    csr = bmat(elements, dtype=dtype).tocsr()

    # when block is squashed in vertical direction
    # we need to swap columns in order to match the order of x, rhs
    # elements

    csr2 = csr[:, np.argsort(np.hstack(map))]
    return csr2


class StrumpackSolver(LinearSolver):
    def __init__(self, gui, engine):
        LinearSolver.__init__(self, gui, engine)

    def spss_set_options(self):
        try:
            import STRUMPACK as ST
        except BaseException:
            assert False, "Can not load STRUMPACK"

        o = getattr(ST, ordering_modes[self.gui.ordering])
        self.spss.set_reordering_method(o)

        if self.gui.mc64job.find('default') == -1:
            job = int(self.gui.mc64job.split('(')[0])
            self.spss.set_matching(job)
        else:
            self.spss.set_matching(ST.STRUMPACK_MATCHING_NONE)

        # compression
        o = getattr(ST, compression_modes[self.gui.compression])
        self.spss.set_compression(o)
        if self.gui.compression_rel_tol.find('default') == -1:
            tol = float(self.gui.compression_rel_tol.split('(')[0])
            self.spss.set_compression_rel_tol(tol)
        if self.gui.compression_abs_tol.find('default') == -1:
            tol = float(self.gui.compression_abs_tol.split('(')[0])
            self.spss.set_compression_abs_tol(tol)
        if self.gui.compression == "HOLBF":
            l = int(self.gui.hodlr_butterfly_levels.split('(')[0])
            self.spss.set_compression_butterfly_levels(l)
        if self.gui.compression_min_sep_size.find('default') == -1:
            l = int(self.gui.compression_min_sep_size.split('(')[0])
            self.spss.set_compression_min_sep_size(l)
        if self.gui.compression_min_front_size.find('default') == -1:
            l = int(self.gui.compression_min_front_size.split('(')[0])
            self.spss.set_compression_min_front_size(l)
        if self.gui.compression_leaf_size.find('default') == -1:
            l = int(self.gui.compression_leaf_size.split('(')[0])
            self.spss.set_compression_leaf_size(l)

        # iterative
        o = getattr(ST, krylov_modes[self.gui.krylov])
        self.spss.set_Krylov_solver(o)
        if self.gui.rctol.find('default') == -1:
            tol = float(self.gui.rctol.split('(')[0])
            self.spss.set_rel_tol(tol)
        if self.gui.actol.find('default') == -1:
            tol = float(self.gui.actol.split('(')[0])
            self.spss.set_abs_tol(tol)
        if self.gui.gmres_restart.find('default') == -1:
            m = int(self.gui.gmres_restart.split('(')[0])
            self.spss.set_gmres_restart(m)
        gs = getattr(ST, GramSchmidt_types[self.gui.gstype])
        self.spss.set_GramSchmidt_type(gs)
        if self.gui.maxiter.find('default') == -1:
            it = int(self.gui.maxiter.split('(')[0])
            self.spss.set_maxit(it)

        # gpu
        if self.gui.use_gpu:
            self.spss.enable_gpu()
        else:
            self.spss.disable_gpu()

    def spss_options_args(self):
        opts = ["--sp_enable_METIS_NodeND", ]
#        opts = ["--sp_enable_METIS_NodeNDP", ]
#               "--sp_enable_METIS_NodeND"]
        if self.gui.lossy_precision.find('default') == -1:
            tol = int(self.gui.lossy_precision.split('(')[0])
            if tol != 16:
                opts.extend(["--sp_lossy_precision", str(tol)])

        if self.gui.use_gpu:
            if self.gui.cuda_cutoff.find('default') == -1:
                cutoff = int(self.gui.cuda_cutoff.split('(')[0])
                opts.extend(["--sp_cuda_cutoff", str(cutoff)])
            if self.gui.cuda_streams.find('default') == -1:
                streams = int(self.gui.cuda_streams.split('(')[0])
                opts.extend(["--sp_cuda_streams", str(streams)])

        if self.gui.MUMPS_SYMQAMD:
            opts.extend(["--sp_enable_MUMPS_SYMQAMD", ])
        else:
            opts.extend(["--sp_disable_MUMPS_SYMQAMD", ])

        if self.gui.agg_amalg:
            opts.extend(["--sp_enable_agg_amalg", ])
        else:
            opts.extend(["--sp_disable_agg_amalg", ])

        if self.gui.indirect_sampling:
            opts.extend(["--sp_enable_indirect_sampling", ])
        else:
            opts.extend(["--sp_disable_indirect_sampling", ])

        if self.gui.replace_tiny_pivots:
            opts.extend(["--sp_enable_replace_tiny_pivots", ])
        else:
            opts.extend(["--sp_disable_replace_tiny_pivots", ])

        if self.gui.separator_ordering_level.find('default') == -1:
            l = int(self.gui.separator_ordering_level.split('(')[0])
            opts.extend(["--sp_separator_ordering_level", str(l)])

        for x in self.gui.extra_options.split("\n"):
            opts.extend([x.strip()
                        for x in x.split(" ") if len(x.strip()) > 0])

        # opts.append("")
        return opts

    def AllocSolver(self, is_complex, use_single_precision):
        try:
            import STRUMPACK as ST
        except BaseException:
            assert False, "Can not load STRUMPACK"
        dprint1("AllocSolver", is_complex, use_single_precision)

        opts = self.spss_options_args()
        dprint1("options", opts, notrim=True)
        verbose = self.gui.log_level > 0
        if use_parallel:
            args = (MPI.COMM_WORLD, opts, verbose)
        else:
            args = (opts, verbose)

        if is_complex:
            if use_single_precision:
                dtype = np.complex64
                if self.gui.use_64_int:
                    spss = ST.C64StrumpackSolver(*args)
                else:
                    spss = ST.CStrumpackSolver(*args)
            else:
                dtype = np.complex128
                if self.gui.use_64_int:
                    spss = ST.Z64StrumpackSolver(*args)
                else:
                    spss = ST.ZStrumpackSolver(*args)
        else:
            if use_single_precision:
                dtype = np.float32
                if self.gui.use_64_int:
                    spss = ST.S64StrumpackSolver(*args)
                else:
                    spss = ST.SStrumpackSolver(*args)
            else:
                dtype = np.float64
                if self.gui.use_64_int:
                    spss = ST.D64StrumpackSolver(*args)
                else:
                    spss = ST.DStrumpackSolver(*args)

        assert spss.isValid(), "Failed to create STRUMPACK solver object"

        spss.set_from_options()

        self.dtype = dtype
        self.spss = spss
        self.is_complex = is_complex
        self.spss_set_options()

    def SetOperator(self, A, dist, name=None):
        try:
            from mpi4py import MPI
        except BaseException:
            from petram.helper.dummy_mpi import MPI
        myid = MPI.COMM_WORLD.rank
        nproc = MPI.COMM_WORLD.size

        self.row_offsets = A.RowOffsets()

        AA = build_csr_local(A, self.dtype, self.is_complex)

        if self.gui.write_mat:
            write_coo_matrix('matrix', AA.tocoo())

        if dist:
            self.spss.set_distributed_csr_matrix(AA)
        else:
            self.spss.set_csr_matrix(AA)
        self._matrix = AA

    def Mult(self, b, x=None, case_base=0):
        try:
            from mpi4py import MPI
        except BaseException:
            from petram.helper.dummy_mpi import MPI
        myid = MPI.COMM_WORLD.rank
        nproc = MPI.COMM_WORLD.size
        try:
            import STRUMPACK as ST
        except BaseException:
            assert False, "Can not load STRUMPACK"

        sol = []
        row_offsets = self.row_offsets.ToList()

        MPI.COMM_WORLD.Barrier()

        dprint1("calling reorder", debug.format_memory_usage())
        ret = self.spss.reorder()

        ret = np.sum(MPI.COMM_WORLD.allgather(int(ret != ST.STRUMPACK_SUCCESS)))
        if ret > 0:
            assert False, "error during recordering (Strumpack)"

        MPI.COMM_WORLD.Barrier()

        dprint1("calling factor", debug.format_memory_usage())
        ret = self.spss.factor()

        zero_pivot = np.sum(MPI.COMM_WORLD.allgather(
            int(ret == ST.STRUMPACK_ZERO_PIVOT)))
        if zero_pivot != 0:
            dprint1("!!!! Zero pivots are detected on " +
                    str(zero_pivot) + " processes")
            dprint1("!!!! continuing ....")

        ret = np.sum(MPI.COMM_WORLD.allgather(int(ret != ST.STRUMPACK_SUCCESS and
                                                  ret != ST.STRUMPACK_ZERO_PIVOT)))
        if ret > 0:
            assert False, "error during factor (Strumpack)"

        return_distributed = self.gui.use_dist_sol

        for kk, bb in enumerate(b):
            rows = MPI.COMM_WORLD.allgather(np.int32(bb.Size()))
            rowstarts = np.hstack((0, np.cumsum(rows)))
            # nicePrint("rowstarts/offser",rowstarts, row_offsets)
            if x is None:
                xx = mfem.BlockVector(self.row_offsets)
                xx.Assign(0.0)
            else:
                xx = x

            if self.is_complex:
                tmp1 = []
                tmp2 = []
                for i in range(len(row_offsets) - 1):
                    bbv = bb.GetBlock(i).GetDataArray()
                    xxv = xx.GetBlock(i).GetDataArray()
                    ll = bbv.size
                    bbv = bbv[:ll // 2] + 1j * bbv[ll // 2:]
                    xxv = xxv[:ll // 2] + 1j * xxv[ll // 2:]
                    tmp1.append(bbv)
                    tmp2.append(xxv)
                bbv = np.hstack(tmp1)
                xxv = np.hstack(tmp2)
            else:
                bbv = bb.GetDataArray()
                xxv = xx.GetDataArray()

            if self.gui.write_mat:
                write_vector('rhs_' + str(kk), bbv)
                write_vector('x_' + str(kk), xxv)

            sys.stdout.flush()
            sys.stderr.flush()

            MPI.COMM_WORLD.Barrier()
            dprint1("calling solve", debug.format_memory_usage())
            ret = self.spss.solve(bbv, xxv, False)

            ret = np.sum(MPI.COMM_WORLD.allgather(
                int(ret != ST.STRUMPACK_SUCCESS)))
            if ret > 0:
                assert False, "error during solve phase (Strumpack)"

            if return_distributed:
                sol.append(xxv)
            else:
                s = []
                for i in range(len(row_offsets) - 1):
                    r1 = row_offsets[i]
                    r2 = row_offsets[i + 1]

                    if self.is_complex:
                        r1 = r1 // 2
                        r2 = r2 // 2
                    xxvv = xxv[r1:r2]

                    if use_parallel:
                        vv = gather_vector(xxvv)
                    else:
                        vv = xxvv.copy()

                    if myid == 0:
                        s.append(vv)
                    else:
                        pass
                if myid == 0:
                    sol.append(np.hstack(s))

        if return_distributed:
            sol = np.transpose(np.vstack(sol))
            return sol
        else:
            if myid == 0:
                sol = np.transpose(np.vstack(sol))
                return sol
            else:
                return None



from __future__ import print_function
from petram.mfem_config import use_parallel
from petram.helper.matrix_file import write_matrix, write_vector
'''
 DoF mapper
    this routine returns M.
     [y_dest] = M [y_src]

     fes1 : source
     fes2 : dest (size of fes2 can be smaller)

     For periodic BC, DoF corresponts to y_dest will be eliminated
     from the final linear system.

        Since, mapping is 
           [1,  0][y_src ]   [y_src]
           [     ][      ] = []
           [M,  0][void  ]   [y_dest]

        A linear system below
             [y_src ]   [b1]
           A [      ] = [  ]
             [y_dest]   [b2]
        becomes
           Pt A P [y_src] = [b1 + Mt b2]

     For H1, L2 element M^-1 = M^t,
     For ND, and RT, M^-1 needs inversion. 
'''

import numpy as np
import scipy
import bisect
import warnings
import traceback
from scipy.sparse import lil_matrix
import petram.debug as debug
debug.debug_default_level = 1
dprint1, dprint2, dprint3 = debug.init_dprints('dof_map')

# this is for testing. need to add three lines below to suppress
# numpy related warning (copied from numpy.__init__.py)
warnings.filterwarnings('error', category=RuntimeWarning)
# Filter out Cython harmless warnings
warnings.filterwarnings("ignore", message="numpy.dtype size changed")
warnings.filterwarnings("ignore", message="numpy.ufunc size changed")
warnings.filterwarnings("ignore", message="numpy.ndarray size changed")


if use_parallel:
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    num_proc = MPI.COMM_WORLD.size
    myid = MPI.COMM_WORLD.rank
    from petram.helper.mpi_recipes import *
    import mfem.par as mfem
    from mfem.common.mpi_debug import nicePrint

else:
    import mfem.ser as mfem
    num_proc = 1
    myid = 0
    def allgather(x): return [x]
    nicePrint = dprint1

mapper_debug = False
gf_debug = False
element_data_debug = False

methods = {}
methods['Dom'] = {'N': 'GetNE',
                  'AttributeArray': 'GetAttributeArray',
                  'Vertices': 'GetElementVertices',
                  'Transformation': 'GetElementTransformation',
                  'Element': 'GetFE',
                  'VDofs': 'GetElementVDofs',
                  'VDofTransformation': 'GetElementVDofTransformation'}
methods['Bdr'] = {'N': 'GetNBE',
                  'AttributeArray': 'GetBdrAttributeArray',
                  'Vertices': 'GetBdrElementVertices',
                  'Transformation': 'GetBdrElementTransformation',
                  'Element': 'GetBE',
                  'VDofs': 'GetBdrElementVDofs',
                  'VDofTransformation': 'GetBdrElementVDofTransformation'}


def notrans(xyz):
    return np.array(xyz, copy=False)


def get_surface_mode(dim, sdim):
    mode1 = ''
    if sdim == 3:
        if dim == 3:
            mode1 = 'Bdr'
        elif dim == 2:
            mode1 = 'Dom'
    elif sdim == 2:
        if dim == 2:
            mode1 = 'Dom'

    if mode1 == '':
        assert False, "not supprint dim/sdim "+str(dim)+'/'+str(sdim)
    return mode1


def get_edge_mode(dim, sdim):
    mode1 = ''
    if sdim == 3:
        if dim == 2:
            mode1 = 'Bdr'
        if dim == 1:
            mode1 = 'Dom'
    elif sdim == 2:
        if dim == 2:
            mode1 = 'Bdr'
        if dim == 1:
            mode1 = 'Dom'
    if mode1 == '':
        assert False, "not supported dim/sdim "+str(dim)+'/'+str(sdim)
    return mode1


def get_volume_mode(dim, sdim):
    if sdim == 3 and dim == 3:
        return 'Dom'
    else:
        assert False, "volume mapping must be 3d mesh"
    return mode1


def find_element(fes, attr, mode='Bdr'):
    mesh = fes.GetMesh()
    m = getattr(mesh, methods[mode]['AttributeArray'])
    arr = m()
    flag = np.in1d(arr, attr)
    return np.arange(len(arr))[flag]


def find_el_center(fes, ibdr1, trans1, mode='Bdr'):

    if len(ibdr1) == 0:
        return np.empty(shape=(0, 2))
    mesh = fes.GetMesh()
    m = getattr(mesh, methods[mode]['Vertices'])
    pts = np.mean(np.array([[trans1(mesh.GetVertexArray(i)) for i in m(k)]
                            for k in ibdr1]), 1)
    # pts = np.vstack([np.mean([trans1(mesh.GetVertexArray(kk))
    #                          for kk in m(k)],0) for k in ibdr1])
    return pts


def get_element_data(fes, idx, trans, mode='Bdr', use_global=False):
    mesh = fes.GetMesh()

    GetTrans = getattr(fes, methods[mode]['Transformation'])
    GetElement = getattr(fes, methods[mode]['Element'])
    GetVDofs = getattr(fes, methods[mode]['VDofs'])

    ret = [None]*len(idx)
    for iii, k1 in enumerate(idx):
        tr1 = GetTrans(k1)
        nodes1 = GetElement(k1).GetNodes()
        vdof1 = GetVDofs(k1)
        # pt1 = np.vstack([trans(tr1.Transform(nodes1.IntPoint(kk)))
        #                 for kk in range(len(vdof1))])
        pt1o = np.vstack([tr1.Transform(nodes1.IntPoint(kk))
                          for kk in range(len(vdof1))])
        pt1 = np.vstack([trans(pt) for pt in pt1o])

        subvdof1 = [x if x >= 0 else -1-x for x in vdof1]
        if use_parallel:
            if use_global:
                subvdof2 = [fes.GetGlobalTDofNumber(i) for i in subvdof1]
            else:
                subvdof2 = [fes.GetLocalTDofNumber(i) for i in subvdof1]
                flag = False
                myoffset = fes.GetMyTDofOffset()
                for k, x in enumerate(subvdof2):
                    if x >= 0:
                        subvdof2[k] = myoffset + x
                    else:
                        flag = True
                if element_data_debug and flag:
                    dprint2(subvdof1, vdof1, subvdof2)
            # note subdof2 = -1 if it is not owned by the node
        else:
            subvdof2 = subvdof1

        newk1 = np.vstack([(k, xx[0], xx[1])
                           for k, xx in enumerate(zip(vdof1, subvdof2))])

        # this doesn't do anything since kk is [0, 1,2,3,...]??
        #pt1 =  np.vstack([pt1[kk] for kk, v, s in newk1])
        #pt1o = np.vstack([pt1o[kk] for kk, v, s in newk1])

        ret[iii] = (newk1, pt1, pt1o)
    return ret


def get_shape(fes, ibdr, mode='Bdr'):
    mesh = fes.GetMesh()

    use_weight = False
    fec_name = fes.FEColl().Name()
    if (fec_name.startswith('RT') and
            mode == 'Bdr'):
        use_weight = True
    if (fec_name.startswith('ND')):
        use_weight = False
        # mode == 'Bdr'):
        use_weight = True
        #use_weight = False

    GetTrans = getattr(fes, methods[mode]['Transformation'])
    GetElement = getattr(fes, methods[mode]['Element'])
    GetVDofs = getattr(fes, methods[mode]['VDofs'])
    GetVDofTrans = getattr(fes, methods[mode]['VDofTransformation'])

    ret = [None]*len(ibdr)
    if len(ibdr) == 0:
        return ret

    # this is to make sure that IntegraitonPoint in Eltrans is
    # set once...
    tr1 = GetTrans(0)
    el = GetElement(0)
    nodes1 = el.GetNodes()
    doftrans = GetVDofTrans(0)
    tr1.SetIntPoint(nodes1.IntPoint(0))

    for iii, k1 in enumerate(ibdr):
        tr1 = GetTrans(k1)
        el = GetElement(k1)
        nodes1 = el.GetNodes()

        weight = tr1.Weight() if use_weight else 1

        v = mfem.Vector(nodes1.GetNPoints())
        shape = [None]*nodes1.GetNPoints()
        for idx in range(len(shape)):
            el.CalcShape(nodes1.IntPoint(idx), v)
            shape[idx] = v.GetDataArray()[idx]*weight
        ret[iii] = np.array(shape)

    return ret


def get_vshape(fes, ibdr, mode='Bdr'):
    mesh = fes.GetMesh()

    GetTrans = getattr(fes, methods[mode]['Transformation'])
    GetElement = getattr(fes, methods[mode]['Element'])
    GetVDofs = getattr(fes, methods[mode]['VDofs'])
    GetVDofTrans = getattr(fes, methods[mode]['VDofTransformation'])

    ret = [None]*len(ibdr)
    if len(ibdr) == 0:
        return ret

    # this is to make sure that IntegraitonPoint in Eltrans is
    # set once...
    tr1 = GetTrans(0)
    el = GetElement(0)
    nodes1 = el.GetNodes()
    doftrans = GetVDofTrans(0)
    tr1.SetIntPoint(nodes1.IntPoint(0))

    v0 = mfem.Vector(tr1.GetSpaceDim())

    use_weight = True
    for iii, k1 in enumerate(ibdr):
        tr1 = GetTrans(k1)
        el = GetElement(k1)
        nodes1 = el.GetNodes()
        doftrans = GetVDofTrans(k1)

        m = mfem.DenseMatrix(nodes1.GetNPoints(),
                             tr1.GetSpaceDim())
        shape = [None]*nodes1.GetNPoints()

        vv = mfem.Vector(nodes1.GetNPoints())

        for idx in range(len(shape)):
            tr1.SetIntPoint(nodes1.IntPoint(idx))
            el.CalcVShape(tr1, m)

            if doftrans is not None:
                vv.Assign(0.0)
                vv[idx] = 1
                doftrans.InvTransformPrimal(vv)
                m.MultTranspose(vv, v0)
                shape[idx] = v0.GetDataArray().copy()
            else:
                shape[idx] = m.GetDataArray()[idx, :].copy()
        ret[iii] = shape
    return ret


def redistribute_pt2_k2(pt2all,  pto2all, k2all, sh2all, map_1_2):
    # map matrix is filled where fes1 elements are owned
    # we deliver pt2 data to nodes where map filling takes place.

    # sort map_1_2 to make bisect work
    sort_idx = np.argsort(map_1_2)
    map_1_2 = np.sort(map_1_2)
    isort_idx = np.arange(len(sort_idx))[np.argsort(sort_idx)]

    k2offset = np.cumsum(comm.allgather(len(k2all)))

    segs = [0]+[bisect.bisect_left(map_1_2, i) for i in k2offset]

    data = [map_1_2[segs[i]:segs[i+1]] for i in range(num_proc)]

    destinations = comm.alltoall(data)
    # nicePrint(destinations)
    k2offset = np.hstack(([0], k2offset))
    myoffset2 = k2offset[myid]

    data1 = []
    data2 = []
    data3 = []
    data4 = []
    for mm in destinations:
        idx = [x-myoffset2 for x in mm]
        data1.append([pt2all[x] for x in idx])
        data2.append([pto2all[x] for x in idx])
        data3.append([k2all[x] for x in idx])
        data4.append([sh2all[x] for x in idx])

    pt2all = sum(comm.alltoall(data1), [])
    pto2all = sum(comm.alltoall(data2), [])
    k2all = sum(comm.alltoall(data3), [])
    sh2all = sum(comm.alltoall(data4), [])

    pt2all = [pt2all[x] for x in isort_idx]
    pto2all = [pto2all[x] for x in isort_idx]
    k2all = [k2all[x] for x in isort_idx]
    sh2all = [sh2all[x] for x in isort_idx]

    map_1_2 = np.arange(len(pt2all))
    # nicePrint(k2all)
    return pt2all,  pto2all, k2all, sh2all, map_1_2


def redistribute_external_entry(external_entry, rstart):
    #  redistribute external entry. first regroup
    #  entries to the destination. then all-to-all

    rstarts = comm.allgather(rstart)
    dests = [bisect.bisect_right(rstarts, r)for r, c, d in external_entry]

    data = [[] for x in range(num_proc)]
    for i, d in zip(dests, external_entry):
        data[i].append(d)

    external_entry = sum(comm.alltoall(data), [])
    return np.array(external_entry)


def map_dof_scalar(map, fes1, fes2, pt1all, pt2all, pto1all, pto2all,
                   k1all, k2all, sh1all, sh2all, map_1_2,
                   trans1, trans2, tol, tdof, rstart):

    dprint1("map_dof_scalar1", debug.format_memory_usage())

    pt = []
    subvdofs1 = []
    subvdofs2 = []

    num_entry = 0
    num_pts = 0

    decimals = int(np.abs(np.log10(tol)))

    if use_parallel:
        P = fes1.Dof_TrueDof_Matrix()
        from mfem.common.parcsr_extra import ToScipyCoo
        P = ToScipyCoo(P).tocsr()
        # this is global TrueDoF (offset is not subtracted)
        VDoFtoGTDoF = P.indices
        external_entry = []
        gtdof_check = []

    for k0 in range(len(pt1all)):
        k2 = map_1_2[k0]
        pt1 = pt1all[k0]
        pto1 = pto1all[k0]
        newk1 = k1all[k0]  # (i local DoF, global DoF)
        sh1 = sh1all[k0]
        pt2 = pt2all[k2]
        pto2 = pto2all[k2]
        newk2 = k2all[k2]
        sh2 = sh2all[k2]
        for k, p in enumerate(pt1):
            num_pts = num_pts + 1

            # if newk1[k,2] in tdof: continue
            iii = bisect.bisect_left(tdof, newk1[k, 2])
            if iii != len(tdof) and tdof[iii] == newk1[k, 2]:
                continue

            # if newk1[k,2] in subvdofs1: continue
            iii = bisect.bisect_left(subvdofs1, newk1[k, 2])
            if iii != len(subvdofs1) and subvdofs1[iii] == newk1[k, 2]:
                continue
            #print("maping points", pt1, pt2)
            dist = np.sum((pt2-p)**2, 1)
            d = np.where(dist == np.min(dist))[0]
            if myid == 1:
                dprint2('min_dist', np.min(dist))

            if len(d) == 1:
                d = d[0]
                s1 = sh1[newk1[k, 0]]
                s2 = sh2[newk2[d, 0]]
                # dprint1("case1 ", s1, s2) #this looks all 1
                if s1/s2 < 0:
                    dprint2("not positive")
                #if myid == 1: print(newk1[d][2]-rstart, newk2[k][2])
                value = np.around(s1/s2, decimals)
                #print("looking into this...", newk1[k], newk2[d])
                if newk1[k, 2] != -1:
                    # flip the sign to match the orientation
                    # need this for mapping ND on edge
                    # we may need to come back here to handle the case
                    # where the edge segment is randomly oriented....
                    if (newk2[d][1]+0.5)*(newk1[k][1]+0.5) < 0:
                        value *= -1
                    map[newk1[k][2]-rstart,
                        newk2[d][2]] = value
                    num_entry = num_entry + 1
                    bisect.insort_left(subvdofs1, newk1[k][2])
                    # subvdofs1.append(newk1[k][2])
                else:
                    # for scalar, this is perhaps not needed
                    # rr = newk1[k][1]] if newk1[k][1]] >= 0 else -1-newk1[k][1]]
                    # gtdof = VDoFtoGTDoF[rr]
                    assert newk1[k][1] >= 0, "Negative index found"
                    gtdof = VDoFtoGTDoF[newk1[k][1]]
                    if not gtdof in gtdof_check:
                        external_entry.append((gtdof, newk2[d][2], value))
                        gtdof_check.append(gtdof)
            else:
                print("failed to map points", pt1, pt2)
                raise AssertionError(
                    "more than two dofs at same places is not supported. ")
        #subvdofs1.extend([s for k, v, s in newk1])
        subvdofs2.extend([s for k, v, s in newk2])

    dprint1("map_dof_scalar2", debug.format_memory_usage())

    if use_parallel:
        dprint1("total entry (before)", sum(allgather(num_entry)))
        external_entry = redistribute_external_entry(
            external_entry, rstart+map.shape[0])

        if len(external_entry.shape) == 2:
            idx1 = np.in1d(external_entry[:, 0], subvdofs1, invert=True)
            val, idx2 = np.unique(external_entry[idx1, 0], return_index=True)
            external_entry = external_entry[idx1][idx2]

            for r, c, d in external_entry:
                # if not r in subvdofs1:
                num_entry = num_entry + 1
                map[r-rstart, c] = d

                #print("adding",myid, r,  c, d )
                # subvdofs1.append(r)

        dprint1("map_dof_scalar3", debug.format_memory_usage())
        '''
        external_entry =  sum(comm.allgather(external_entry),[])           
        for r, c, d in external_entry:
           h = map.shape[0]
           if (r - rstart >= 0 and r - rstart < h and
               not r  in subvdofs1):
               num_entry = num_entry + 1                                 
               print("adding",myid, r,  c, d )
               map[r-rstart, c] = d
               subvdofs1.append(r)
        '''
        total_entry = sum(allgather(num_entry))
        total_pts = sum(allgather(num_pts))
        if sum(allgather(map.nnz)) != total_entry:
            assert False, "total_entry does not match with nnz"
    else:
        total_entry = num_entry
        total_pts = num_pts

    #dprint1("map size", map.shape)
    dprint1("local pts/entry", num_pts, " ", num_entry)
    dprint1("total pts/entry", total_pts, " ", total_entry)
    return map


def map_dof_vector(map, fes1, fes2, pt1all, pt2all, pto1all, pto2all,
                   k1all, k2all, sh1all, sh2all, map_1_2,
                   trans1, trans2, tol, tdof, rstart,
                   old_mapping=True):

    dprint1("map_dof_vector1", debug.format_memory_usage())

    pt = []
    subvdofs1 = []
    subvdofs2 = []

    num1 = 0
    num2 = 0
    num_pts = 0

    decimals = int(np.abs(np.log10(tol)))

    if use_parallel:
        P = fes1.Dof_TrueDof_Matrix()
        from mfem.common.parcsr_extra import ToScipyCoo
        P1mat = ToScipyCoo(P).tocsr()
        # this is global TrueDoF (offset is not subtracted)
        external_entry = []
        gtdof_check = []

    def make_entry(r, c, value, num_entry):
        value = np.around(value, decimals)
        if value == 0:
            return num_entry
        if r[1] != -1:
            map[r[1]-rstart, c] = value
            num_entry = num_entry + 1
            bisect.insort_left(subvdofs1, r[1])
            # subvdofs1.append(r[1])
        else:
            rr = r[0] if r[0] >= 0 else -1-r[0]
            gtdofs = P1mat.indices[P1mat.indptr[rr]:P1mat.indptr[rr+1]]
            weights = P1mat.data[P1mat.indptr[rr]:P1mat.indptr[rr+1]]
            for gtdof, w in zip(gtdofs, weights):
                if not gtdof in gtdof_check:
                    external_entry.append((gtdof, c, value*w))
                    gtdof_check.append(gtdof)

        return num_entry

    tdof = sorted(tdof)

    for k0 in range(len(pt1all)):
        k2 = map_1_2[k0]
        pt1 = pt1all[k0]
        pto1 = pto1all[k0]
        newk1 = k1all[k0]  # (i local DoF, global DoF)
        sh1 = sh1all[k0]
        pt2 = pt2all[k2]
        pto2 = pto2all[k2]
        newk2 = k2all[k2]
        sh2 = sh2all[k2]

        #dprint1(len(np.unique(newk1[:,2])) == len(newk1[:,2]))
        for k, p in enumerate(pt1):
            # if idx[k]: continue
            num_pts = num_pts + 1

            # if newk1[k,2] in tdof: continue
            iii = bisect.bisect_left(tdof, newk1[k, 2])
            if iii != len(tdof) and tdof[iii] == newk1[k, 2]:
                continue

            # if newk1[k,2] in subvdofs1: continue
            iii = bisect.bisect_left(subvdofs1, newk1[k, 2])
            if iii != len(subvdofs1) and subvdofs1[iii] == newk1[k, 2]:
                continue

            dist = np.sum((pt2-p)**2, 1)
            d = np.where(dist == np.min(dist))[0]

            #if myid == 1: dprint1('min_dist', np.min(dist))
            if len(d) == 1:
                '''
                this factor is not always 1
                '''

                d = d[0]
                s = np.sign(newk1[k, 1] + 0.5)*np.sign(newk2[d, 1] + 0.5)
                p1 = pto1[k]
                p2 = pto2[d]

                if np.sum(np.std(sh1, 0)) != 0:
                    delta = np.sum(np.std(pto1, 0))/np.sum(np.std(sh1, 0))/10.
                else:
                    delta = 1  # this happens on edge....

                v1 = trans1(p1) - trans1(p1 + delta*sh1[newk1[k, 0]])
                v2 = trans2(p2) - trans2(p2 + delta*sh2[newk2[d, 0]])

                fac = np.sum(v1*v2)/np.sum(v1*v1)*s
                # except RuntimeWarning:
                #   print(pto1, pto1.shape, p1,p2, s, delta, sh1[newk1[k, 0]], sh2[newk2[d, 0]])
                #   assert False, "Got Here"

                num1 = make_entry(newk1[k, [1, 2]], newk2[d, 2], fac, num1)

            elif len(d) == 2:
                dd = np.argsort(np.sum((pt1 - p)**2, 1))

                p1 = pto1[dd[0]]
                p3 = pto2[d[0]]
                p2 = pto1[dd[1]]
                p4 = pto2[d[1]]
                delta = np.sum(np.std(pto1, 0))/np.sum(np.std(sh1, 0))/10.

                v1 = trans1(p1) - trans1(p1 + delta*sh1[newk1[dd[0], 0]])
                v2 = trans1(p2) - trans1(p2 + delta*sh1[newk1[dd[1], 0]])
                v3 = trans2(p3) - trans2(p3 + delta*sh2[newk2[d[0], 0]])
                v4 = trans2(p4) - trans2(p4 + delta*sh2[newk2[d[1], 0]])
                v1 = v1*np.sign(newk1[dd[0], 1] + 0.5)
                v2 = v2*np.sign(newk1[dd[1], 1] + 0.5)
                v3 = v3*np.sign(newk2[d[0], 1] + 0.5)
                v4 = v4*np.sign(newk2[d[1], 1] + 0.5)
                s = np.sign(newk1[k, 1] + 0.5)*np.sign(newk2[d, 1] + 0.5)

                def vnorm(v):
                    return v/np.sqrt(np.sum(v**2))
                v1n = vnorm(v1)
                v2n = vnorm(v2)
                v3n = vnorm(v3)
                v4n = vnorm(v4)

                if (np.abs(np.abs(np.sum(v1n*v3n))-1) < tol and
                        np.abs(np.abs(np.sum(v2n*v4n))-1) < tol):
                    fac1 = np.sum(v1*v3)/np.sum(v1*v1)
                    fac2 = np.sum(v2*v4)/np.sum(v2*v2)

                    num2 = make_entry(
                        newk1[dd[0], [1, 2]], newk2[d[0], 2], fac1, num2)
                    num2 = make_entry(
                        newk1[dd[1], [1, 2]], newk2[d[1], 2], fac2, num2)

                elif (np.abs(np.abs(np.sum(v2n*v3n))-1) < tol and
                      np.abs(np.abs(np.sum(v1n*v4n))-1) < tol):
                    fac1 = np.sum(v1*v4)/np.sum(v1*v1)
                    fac2 = np.sum(v2*v3)/np.sum(v2*v2)

                    num2 = make_entry(
                        newk1[dd[0], [1, 2]], newk2[d[1], 2], fac1, num2)
                    num2 = make_entry(
                        newk1[dd[1], [1, 2]], newk2[d[0], 2], fac2, num2)
                else:
                    def proj2d(v, e1, e2):
                        return np.array([np.sum(v*e1), np.sum(v*e2)])
                    if len(v1) == 3:  # if vector is 3D, needs to prjoect on surface
                        e3 = np.cross(v1n, v2n)
                        e1 = v1n
                        e2 = np.cross(e1, e3)
                        v1 = proj2d(v1, e1, e2)
                        v2 = proj2d(v2, e1, e2)
                        v3 = proj2d(v3, e1, e2)
                        v4 = proj2d(v4, e1, e2)

                    if old_mapping:
                        m1 = np.transpose(np.vstack((v1, v2)))
                        m2 = np.transpose(np.vstack((v3, v4)))
                        m = np.dot(np.linalg.inv(m1), m2)
                        m = np.around(np.linalg.inv(m), decimals=decimals)
                        m = np.transpose(m)
                    else:
                        m1 = np.vstack((v1, v2))
                        m2 = np.vstack((v3, v4))
                        m = np.dot(m2, np.linalg.inv(m1))
                        m = np.around(m, decimals=decimals)
                        m = np.transpose(m)
                    '''
                   #this was wrong. keeping it for now for record (2020.03)
                   '''
                    num2 = make_entry(
                        newk1[dd[0], [1, 2]], newk2[d[0], 2], m[0, 0], num2)
                    num2 = make_entry(
                        newk1[dd[0], [1, 2]], newk2[d[1], 2], m[0, 1], num2)
                    num2 = make_entry(
                        newk1[dd[1], [1, 2]], newk2[d[0], 2], m[1, 0], num2)
                    num2 = make_entry(
                        newk1[dd[1], [1, 2]], newk2[d[1], 2], m[1, 1], num2)

            elif len(d) == 3:
                dd = np.argsort(np.sum((pt1 - p)**2, 1))

                p1 = [pto1[dd[i]] for i in [0, 1, 2]]
                p2 = [pto2[d[i]] for i in [0, 1, 2]]

                delta = np.sum(np.std(pto1, 0))/np.sum(np.std(sh1, 0))/10.

                v1 = [trans1(p1[i]) - trans1(p1[i] + delta*sh1[newk1[dd[i], 0]])
                      for i in [0, 1, 2]]
                v2 = [trans2(p2[i]) - trans2(p2[i] + delta*sh2[newk2[d[i], 0]])
                      for i in [0, 1, 2]]

                v1 = [v1[i]*np.sign(newk1[dd[i], 1] + 0.5) for i in [0, 1, 2]]
                v2 = [v2[i]*np.sign(newk2[d[i], 1] + 0.5) for i in [0, 1, 2]]

                s = np.sign(newk1[k, 1] + 0.5)*np.sign(newk2[d, 1] + 0.5)

                def vnorm(v):
                    return v/np.sqrt(np.sum(v**2))
                v1n = [vnorm(v) for v in v1]
                v2n = [vnorm(v) for v in v2]

                '''
                note: see old_mapping flag note below

                m1 = np.transpose(np.vstack(v1))
                m2 = np.transpose(np.vstack(v2))
                m = np.dot(np.linalg.inv(m1), m2)
                m = np.around(np.linalg.inv(m), decimals = decimals)
                num2 = make_entry(newk1[dd[0],[1,2]], newk2[d[0],2], m[0,0], num2)
                num2 = make_entry(newk1[dd[0],[1,2]], newk2[d[1],2], m[1,0], num2)
                num2 = make_entry(newk1[dd[0],[1,2]], newk2[d[2],2], m[2,0], num2)
                                  
                num2 = make_entry(newk1[dd[1],[1,2]], newk2[d[0],2], m[0,1], num2)
                num2 = make_entry(newk1[dd[1],[1,2]], newk2[d[1],2], m[1,1], num2)
                num2 = make_entry(newk1[dd[1],[1,2]], newk2[d[2],2], m[2,1], num2)
                                  
                num2 = make_entry(newk1[dd[2],[1,2]], newk2[d[0],2], m[0,2], num2)
                num2 = make_entry(newk1[dd[2],[1,2]], newk2[d[1],2], m[1,2], num2)
                num2 = make_entry(newk1[dd[2],[1,2]], newk2[d[2],2], m[2,2], num2)
                '''
                if old_mapping:
                    m1 = np.transpose(np.vstack(v1))
                    m2 = np.transpose(np.vstack(v2))
                    m = np.dot(np.linalg.inv(m1), m2)
                    m = np.around(np.linalg.inv(m), decimals=decimals)
                    m = np.transpose(m)
                else:
                    m1 = np.vstack(v1)
                    m2 = np.vstack(v2)
                    m = np.dot(m2, np.linalg.inv(m1))
                    m = np.around(m, decimals=decimals)
                    m = np.transpose(m)

                num2 = make_entry(newk1[dd[0], [1, 2]],
                                  newk2[d[0], 2], m[0, 0], num2)
                num2 = make_entry(newk1[dd[0], [1, 2]],
                                  newk2[d[1], 2], m[0, 1], num2)
                num2 = make_entry(newk1[dd[0], [1, 2]],
                                  newk2[d[2], 2], m[0, 2], num2)

                num2 = make_entry(newk1[dd[1], [1, 2]],
                                  newk2[d[0], 2], m[1, 0], num2)
                num2 = make_entry(newk1[dd[1], [1, 2]],
                                  newk2[d[1], 2], m[1, 1], num2)
                num2 = make_entry(newk1[dd[1], [1, 2]],
                                  newk2[d[2], 2], m[1, 2], num2)

                num2 = make_entry(newk1[dd[2], [1, 2]],
                                  newk2[d[0], 2], m[2, 0], num2)
                num2 = make_entry(newk1[dd[2], [1, 2]],
                                  newk2[d[1], 2], m[2, 1], num2)
                num2 = make_entry(newk1[dd[2], [1, 2]],
                                  newk2[d[2], 2], m[2, 2], num2)

            else:
                print(pt1, pt2)
                '''
                 newk1 = k1all[k0] #(i local DoF, global DoF)
                 sh1 = sh1all[k0]           
                 pto2 = pto2all[k2]
                 newk2 = k2all[k2]
                 sh2 = sh2all[k2]
                '''
                # to do support three vectors
                raise AssertionError("more than three dofs at same place")
        subvdofs2.extend([s for k, v, s in newk2])

    dprint1("map_dof_vector2", debug.format_memory_usage())
    num_entry = num1 + num2

    if use_parallel:

        dprint1("total entry (before)", sum(allgather(num_entry)))

        #nicePrint("data to exchange", len(external_entry))
        external_entry = redistribute_external_entry(
            external_entry, rstart+map.shape[0])

        if len(external_entry.shape) == 2:
            idx1 = np.in1d(external_entry[:, 0], subvdofs1, invert=True)
            val, idx2 = np.unique(external_entry[idx1, 0], return_index=True)
            external_entry = external_entry[idx1][idx2]

            for r, c, d in external_entry:
                # if not r in subvdofs1:
                num_entry = num_entry + 1
                #print("adding",myid, r,  c, d )
                map[r-rstart, c] = d
                # subvdofs1.append(r)

        dprint1("map_dof_vector3", debug.format_memory_usage())
        '''        
        external_entry =  sum(comm.allgather(external_entry),[])
        #nicePrint(external_entry)        
        for r, c, d in external_entry:
           h = map.shape[0]
           if (r - rstart >= 0 and r - rstart < h and
               not r  in subvdofs1):
               num_entry = num_entry + 1                                 
               print("adding",myid, r,  c, d )
               map[r-rstart, c] = d
               subvdofs1.append(r)
        '''
        total_entry = sum(allgather(num_entry))
        total_pts = sum(allgather(num_pts))
        if sum(allgather(map.nnz)) != total_entry:
            assert False, "total_entry does not match with nnz"
    else:
        total_entry = num_entry
        total_pts = num_pts

    #dprint1("map size", map.shape)
    dprint1("local pts/entry", num_pts, " ", num_entry)
    dprint1("total pts/entry", total_pts, " ", total_entry)

    return map


def gather_dataset(idx1, idx2, fes1, fes2, trans1,
                   trans2, tol, shape_type='scalar',
                   mode='surface'):

    #dprint1("gather_dataset1", debug.format_memory_usage())

    if fes2 is None:
        fes2 = fes1
    if trans1 is None:
        trans1 = notrans
    if trans2 is None:
        trans2 = notrans

    mesh1 = fes1.GetMesh()
    mesh2 = fes2.GetMesh()

    if mode == 'volume':
        mode1 = get_volume_mode(mesh1.Dimension(), mesh1.SpaceDimension())
        mode2 = get_volume_mode(mesh2.Dimension(), mesh2.SpaceDimension())
    elif mode == 'surface':
        mode1 = get_surface_mode(mesh1.Dimension(), mesh1.SpaceDimension())
        mode2 = get_surface_mode(mesh2.Dimension(), mesh2.SpaceDimension())
    elif mode == 'edge':
        mode1 = get_edge_mode(mesh1.Dimension(), mesh1.SpaceDimension())
        mode2 = get_edge_mode(mesh2.Dimension(), mesh2.SpaceDimension())
    elif mode == 'edge':
        mode1 = get_point_mode(mesh1.Dimension(), mesh1.SpaceDimension())
        mode2 = get_point_mode(mesh2.Dimension(), mesh2.SpaceDimension())

    # collect data
    ibdr1 = find_element(fes1, idx1, mode=mode1)
    ibdr2 = find_element(fes2, idx2, mode=mode2)

    ct1 = find_el_center(fes1, ibdr1, trans1, mode=mode1)
    ct2 = find_el_center(fes2, ibdr2, trans2, mode=mode2)
    arr1 = get_element_data(fes1, ibdr1, trans1, mode=mode1)
    arr2 = get_element_data(fes2, ibdr2, trans1, mode=mode2, use_global=True)

    if shape_type == 'scalar':
        sh1all = get_shape(fes1, ibdr1, mode=mode1)
        sh2all = get_shape(fes2, ibdr2, mode=mode2)
    elif shape_type == 'vector':
        sh1all = get_vshape(fes1, ibdr1, mode=mode1)
        sh2all = get_vshape(fes2, ibdr2, mode=mode2)
    else:
        assert False, "Unknown shape type"

    #dprint1("gather_dataset2", debug.format_memory_usage())

    # pt is on (u, v), pto is (x, y, z)
    try:
        k1all, pt1all, pto1all = zip(*arr1)
    except:
        k1all, pt1all, pto1all = (), (), ()
    try:
        k2all, pt2all, pto2all = zip(*arr2)
    except:
        k2all, pt2all, pto2all = (), (), ()

    if use_parallel:
        # share ibr2 (destination information among nodes...)
        ct1dim = ct1.shape[1] if ct1.size > 0 else 0
        ct1dim = comm.allgather(ct1dim)
        ct1 = np.atleast_2d(ct1).reshape(-1, max(ct1dim))
        ct2 = np.atleast_2d(ct2).reshape(-1, max(ct1dim))
        ct2 = allgather_vector(ct2, MPI.DOUBLE)

    #dprint1("gather_dataset3", debug.format_memory_usage())

    # mapping between elements
    from scipy.spatial import cKDTree
    tree = cKDTree(ct2)
    ctr_dist, map_1_2 = tree.query(ct1)

    #dprint1("gather_dataset4", debug.format_memory_usage())

    if ctr_dist.size > 0 and np.max(ctr_dist) > 1e-15:
        print('Center Dist may be too large (check mesh): ' +
              str(np.max(ctr_dist)))

    if use_parallel:

        pt2all, pto2all, k2all, sh2all, map_1_2 = redistribute_pt2_k2(pt2all,
                                                                      pto2all, k2all, sh2all,
                                                                      map_1_2)

        #dprint1("gather_dataset5", debug.format_memory_usage())

    # map is fill as transposed shape (row = fes1)

    data = pt1all, pt2all, pto1all, pto2all, k1all, k2all, sh1all, sh2all,

    return data, map_1_2


def get_empty_map(fes1, fes2, is_complex=False):
    '''
    empty matrix (fes1: test, fes2: 
    '''
    if use_parallel:
        fesize1 = fes1.GetTrueVSize()
        fesize2 = fes2.GlobalTrueVSize()
        rstart = fes1.GetMyTDofOffset()
    else:
        fesize1 = fes1.GetNDofs()
        fesize2 = fes2.GetNDofs()
        rstart = 0

    dtype = complex if is_complex else float
    map = lil_matrix((fesize1, fesize2), dtype=dtype)
    return map, rstart


def map_xxx_h1(xxx, idx1, idx2, fes1, fes2=None, trans1=None,
               trans2=None, tdof1=None, tdof2=None, tol=1e-4,
               old_mapping=True):
    '''
    map DoF on surface to surface

      fes1: source finite element space
      fes2: destination finite element space

      idx1: surface attribute (Bdr for 3D/3D, Domain for 2D/3D or 2D/2D)

    '''

    if fes2 is None:
        fes2 = fes1
    if trans1 is None:
        trans1 = notrans
    if trans2 is None:
        trans2 = trans1
    if tdof1 is None:
        tdof1 = []
    if tdof2 is None:
        tdof2 = []

    tdof = tdof1  # ToDo support tdof2
    map, rstart = get_empty_map(fes1, fes2)
    data, elmap = gather_dataset(idx1, idx2, fes1, fes2, trans1,
                                 trans2, tol, shape_type='scalar',
                                 mode=xxx)

    pt1all, pt2all, pto1all, pto2all, k1all, k2all, sh1all, sh2all = data

    map_dof_scalar(map, fes1, fes2, pt1all, pt2all, pto1all, pto2all,
                   k1all, k2all, sh1all, sh2all, elmap,
                   trans1, trans2, tol, tdof1, rstart)

    return map


def map_volume_h1(*args, **kwargs):
    return map_xxx_h1('volume', *args, **kwargs)


def map_surface_h1(*args, **kwargs):
    return map_xxx_h1('surface', *args, **kwargs)


def map_edge_h1(*args, **kwargs):
    return map_xxx_h1('edge', *args, **kwargs)


def map_point_h1(idx1, idx2, fes1, fes2=None, trans1=None,
                 trans2=None, tdof1=None, tdof2=None, tol=1e-4,
                 old_mapping=True):
    '''
    map DoF from vertex to vertex

      fes1: source finite element space
      fes2: destination finite element space

      idx1: surface attribute (Bdr for 3D/3D, Domain for 2D/3D or 2D/2D)

    '''

    if fes2 is None:
        fes2 = fes1
    if trans1 is None:
        trans1 = notrans
    if trans2 is None:
        trans2 = trans1
    if tdof1 is None:
        tdof1 = []
    if tdof2 is None:
        tdof2 = []

    tdof = tdof1  # ToDo support tdof2
    map, rstart = get_empty_map(fes1, fes2)

    R = fes1.GetConformingRestriction()
    if R is not None:
        assert False, "Non-conforming mesh not supported"
    '''
    If R is not none. we have to do something like this...
        VDof2TDof = np.zeros(fes.GetNDofs(), dtype=int)
        for i, j in enumerate(R.GetJArray()):
            VDof2TDof[j] = i
        TDof2Vdof = R.GetJArray().copy()            
    '''
    mesh = fes1.GetMesh()
    battrs = {battr: k for k, battr in enumerate(mesh.GetBdrAttributeArray())}

    vdofs1 = [fes1.GetBdrElementVDofs(
        battrs[i])[0] if i in battrs else -1 for i in idx1]
    vdofs2 = [fes2.GetBdrElementVDofs(
        battrs[i])[0] if i in battrs else -1 for i in idx2]

    if use_parallel:
        P = fes1.Dof_TrueDof_Matrix()
        from mfem.common.parcsr_extra import ToScipyCoo
        P = ToScipyCoo(P).tocsr()

        # this is global TrueDoF (offset is not subtracted # this is okay for scalar)
        VDoFtoGTDoF = P.indices
        gtdofs1 = [VDoFtoGTDoF[x] if x >= 0 else -1 for x in vdofs1]
        gtdofs2 = [VDoFtoGTDoF[x] if x >= 0 else -1 for x in vdofs2]

        # collecting non-negative tdofs2
        gtdofs2 = allgather_vector(np.array(gtdofs2, dtype=int))
        tmp = gtdofs2.reshape(-1, len(idx2))
        gtdofs2 = [np.max(tmp[:, i]) for i in range(len(idx2))]

    else:
        gtdofs1 = vdofs1
        gtdofs2 = vdofs2

    for v1, v2 in zip(gtdofs1, gtdofs2):
        if v1 <= 0:
            continue
        map[v1-rstart, v2] = 1
    return map


def map_xxx_nd(xxx, idx1, idx2, fes1, fes2=None, trans1=None,
               trans2=None, tdof1=None, tdof2=None, tol=1e-4,
               old_mapping=True):
    '''
    map DoF on surface to surface

      fes1: source finite element space
      fes2: destination finite element space

      idx1: surface attribute (Bdr for 3D/3D, Domain for 2D/3D or 2D/2D)

    '''

    if fes2 is None:
        fes2 = fes1
    if trans1 is None:
        trans1 = notrans
    if trans2 is None:
        trans2 = trans1
    if tdof1 is None:
        tdof1 = []
    if tdof2 is None:
        tdof2 = []

    tdof = tdof1  # ToDo support tdof2
    map, rstart = get_empty_map(fes1, fes2)
    data, elmap = gather_dataset(idx1, idx2, fes1, fes2, trans1,
                                 trans2, tol, shape_type='vector',
                                 mode=xxx)

    pt1all, pt2all, pto1all, pto2all, k1all, k2all, sh1all, sh2all = data

    map_dof_vector(map, fes1, fes2, pt1all, pt2all, pto1all, pto2all,
                   k1all, k2all, sh1all, sh2all, elmap,
                   trans1, trans2, tol, tdof1, rstart, old_mapping=old_mapping)

    return map


def map_volume_nd(*args, **kwargs):
    return map_xxx_nd('volume', *args, **kwargs)


def map_surface_nd(*args, **kwargs):
    return map_xxx_nd('surface', *args, **kwargs)


def map_edge_nd(*args, **kwargs):
    return map_xxx_nd('edge', *args, **kwargs)


def map_xxx_rt(xxx, idx1, idx2, fes1, fes2=None, trans1=None,
               trans2=None, tdof1=None, tdof2=None, tol=1e-4,
               old_mapping=True):
    '''
    map DoF on surface to surface

      fes1: source finite element space (row)
      fes2: destination finite element space (col)

      idx1: surface attribute (Bdr for 3D/3D, Domain for 2D/3D or 2D/2D)

    '''

    if fes2 is None:
        fes2 = fes1
    if trans1 is None:
        trans1 = notrans
    if trans2 is None:
        trans2 = trans1
    if tdof1 is None:
        tdof1 = []
    if tdof2 is None:
        tdof2 = []

    map, rstart = get_empty_map(fes1, fes2)
    if xxx == 'volume':
        tdof = tdof1  # ToDo support tdof2
        data, elmap = gather_dataset(idx1, idx2, fes1, fes2, trans1,
                                     trans2, tol, shape_type='vector',
                                     mode=xxx)
        pt1all, pt2all, pto1all, pto2all, k1all, k2all, sh1all, sh2all = data
        map_dof_vector(map, fes1, fes2, pt1all, pt2all, pto1all, pto2all,
                       k1all, k2all, sh1all, sh2all, elmap,
                       trans1, trans2, tol, tdof1, rstart,
                       old_mapping=old_mapping)
    else:
        tdof = tdof1  # ToDo support tdof2
        data, elmap = gather_dataset(idx1, idx2, fes1, fes2, trans1,
                                     trans2, tol, shape_type='scalar',
                                     mode=xxx)

        pt1all, pt2all, pto1all, pto2all, k1all, k2all, sh1all, sh2all = data

        map_dof_scalar(map, fes1, fes2, pt1all, pt2all, pto1all, pto2all,
                       k1all, k2all, sh1all, sh2all, elmap,
                       trans1, trans2, tol, tdof1, rstart)

    return map


def map_volume_rt(*args, **kwargs):
    return map_xxx_rt('volume', *args, **kwargs)


def map_surface_rt(*args, **kwargs):
    return map_xxx_rt('surface', *args, **kwargs)


def projection_matrix(idx1,  idx2,  fes, tdof1, fes2=None, tdof2=None,
                      trans1=None, trans2=None, dphase=0.0, weight=None,
                      tol=1e-7, mode='surface', filldiag=True,
                      old_mapping=True):
    '''
     map: destinatiom mapping 
     smap: source mapping

     old_mapping: True : periodic boundary conditions are implemented this way
                         x = M*y
                  False: projection operator should use this flag.
                         y = M*x
        the difference is only when mapping two/three DoFs sitting at the
        same location. therefore, it only matters for ND and RT cases

    '''
    fec_name = fes.FEColl().Name()
    dprint1("constructing mapping to fec_name", fec_name, mode)
    if fec_name.startswith('ND') and mode == 'volume':
        mapper = map_volume_nd
    elif fec_name.startswith('ND') and mode == 'surface':
        mapper = map_surface_nd
    elif fec_name.startswith('ND') and mode == 'edge':
        mapper = map_edge_nd
    elif fec_name.startswith('H1') and mode == 'volume':
        mapper = map_volume_h1
    elif fec_name.startswith('H1') and mode == 'surface':
        mapper = map_surface_h1
    elif fec_name.startswith('H1') and mode == 'edge':
        mapper = map_edge_h1
    elif fec_name.startswith('H1') and mode == 'point':
        mapper = map_point_h1
    elif fec_name.startswith('L2') and mode == 'volume':
        mapper = map_volume_h1
    elif fec_name.startswith('L2') and mode == 'surface':
        mapper = map_surface_h1
    elif fec_name.startswith('L2') and mode == 'edge':
        mapper = map_edge_h1
    elif fec_name.startswith('RT') and mode == 'volume':
        mapper = map_volume_rt
    elif fec_name.startswith('RT') and mode == 'surface':
        mapper = map_surface_rt
    else:
        raise NotImplementedError("mapping :" + fec_name + ", mode: " + mode)

    map = mapper(idx2, idx1, fes, fes2=fes2, trans1=trans1, trans2=trans2, tdof1=tdof1,
                 tdof2=tdof2, tol=tol, old_mapping=old_mapping)

    if weight is None:
        iscomplex = False
        if (dphase == 0.):
            pass
        elif (dphase == 180.):
            map = map.tocsr()
            map *= -1
            map = map.tolil()

        else:
            iscomplex = True
            map = map.astype(complex)
            # need to this to make efficient....
            map = map.tocsr()
            map *= np.exp(-1j*np.pi/180*dphase)
            map = map.tolil()

    else:
        iscomplex = np.iscomplexobj(weight)
        if iscomplex:
            map = map.astype(complex)
        if map.nnz > 0:
            map = map.tocsr()
            map *= -weight
            map = map.tolil()

    m_coo = map.tocoo()
    row = m_coo.row
    col = m_coo.col
    col = np.unique(col)

    if use_parallel:
        start_row = fes.GetMyTDofOffset()
        end_row = fes.GetMyTDofOffset() + fes.GetTrueVSize()
        col = np.unique(allgather_vector(col))
        row = row + start_row
    else:
        start_row = 0
        end_row = map.shape[0]

    if filldiag:
        for i in range(min(map.shape[0], map.shape[1])):
            r = start_row+i
            if not r in col:
                map[i, r] = 1.0

    from scipy.sparse import coo_matrix, csr_matrix
    if use_parallel:
        if iscomplex:
            m1 = csr_matrix(map.real, dtype=float)
            m2 = csr_matrix(map.imag, dtype=float)
        else:
            m1 = csr_matrix(map.real, dtype=float)
            m2 = None
        from mfem.common.chypre import CHypreMat
        start_col = fes2.GetMyTDofOffset()
        end_col = fes2.GetMyTDofOffset() + fes2.GetTrueVSize()
        col_starts = [start_col, end_col, map.shape[1]]
        M = CHypreMat(m1, m2, col_starts=col_starts)
    else:
        from petram.helper.block_matrix import convert_to_ScipyCoo

        M = convert_to_ScipyCoo(coo_matrix(map, dtype=map.dtype))

    return M, row, col



import os

import wx
import wx.adv
import wx.propgrid as wxpg

from ifigure.utils.edit_list import DialogEditList
from ifigure.ifigure_config import icondir

bmp1=wx.Bitmap(os.path.join(icondir, '16x16', 'smode.png'))
bmp2=wx.Bitmap(os.path.join(icondir, '16x16', 'pmode.png'))

HypreSmoother = ["None", "Jacobi", "l1Jacobi", "l1GS", "l1GStr", 
                "lumpedJacobi", "GS", "Chebyshev", "Taubin",
                 "FIR", "MUMPS"]
SparseSmoother = ["None", "Jacobi", "l1Jacobi", "lumpedJacobi",
                  "GS", "forwardGS", "backwrdGS", "MUMPS"]

class SmootherChoiceProperty(wxpg.StringProperty):
    def __init__(self, label, name=wxpg.PG_LABEL, value='', 
                 choices = None):
        wxpg.StringProperty.__init__(self, label, name, value)
        if choices is None: choices = ["empty choices"]
        # Prepare choices
        dialog_choices = []
        for x in choices:
            dialog_choices.append(x)
        self.dialog_choices = dialog_choices

    def DoGetEditorClass(self):
        return wxpg.PropertyGridInterface.GetEditorByName("TextCtrlAndButton")

    def GetEditorDialog(self):
        # Set what happens on button click
        return SingleChoiceDialogAdapter(self.dialog_choices)

class SingleChoiceDialogAdapter(wxpg.PGEditorDialogAdapter):
    """ This demonstrates use of wxpg.PGEditorDialogAdapter.
    """
    def __init__(self, choices):
        wxpg.PGEditorDialogAdapter.__init__(self)
        self.choices = choices

    def DoShowDialog(self, propGrid, property):
        s = wx.GetSingleChoice("Choose Smoother", "Smoother...", self.choices)

        if s:
            self.SetValue(s)
            return True

        return False;

class WidgetSmoother(wx.Panel):
    def __init__(self, parent, id, *args, **kwargs):

        wx.Panel.__init__(self, parent, id)
        self.prop_names = []
        setting = kwargs.pop("setting", {})

        sizer1 = wx.BoxSizer(wx.HORIZONTAL)
        sizer2 = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer1)
        sizer1.Add(sizer2, 1, wx.EXPAND)
        self.pg = wxpg.PropertyGridManager(self,
                        style=wxpg.PG_SPLITTER_AUTO_CENTER |
                              #wxpg.PG_AUTO_SORT|
                              #wxpg.PG_DESCRIPTION |
                              wxpg.PG_TOOLBAR)
        self.pg.AddPage("Serial", bmp=bmp1)
        self.pg.SelectPage(0)
        self.pg.SetColumnCount(2)
        self.pg.SetColumnTitle(0, "FES var.")
        self.pg.Append(SmootherChoiceProperty("E1", choices=SparseSmoother))

        self.pg.AddPage( "Parallel", bmp=bmp2)
        self.pg.SelectPage(1)
        self.pg.SetColumnCount(2)

        self.pg.SetColumnTitle(0, "FES var.")

        self.pg.Append(SmootherChoiceProperty("E1", choices = HypreSmoother))

        #self.pg.SetPageSplitterPosition(0, 0.5) 
        #self.pg.SetPageSplitterPosition(1, 0.5) 

        self.pg.ShowHeader(True)
        sizer2.Add(self.pg, 1, wx.EXPAND|wx.ALL, 1)

        self.pg.Bind( wxpg.EVT_PG_PAGE_CHANGED, self.OnPropGridPageChange )
        self.pg.Bind( wxpg.EVT_PG_CHANGED, self.OnPropGridChange )
        
        value = [('E1', ['GS', 'GS'],),
                 ('V1', ['GS', 'Jacobi'])]

        self.SetValue(value)

    def set_column_title(self):
        index = self.pg.GetSelectedPage()
        if index == 0:
            self.pg.SetColumnTitle(1, "Serial Preconditioner")
        else:
            self.pg.SetColumnTitle(1, "Parallel Preconditioner")

    def OnPropGridPageChange(self, evt):
        self.set_column_title()
        
    def OnPropGridChange(self, evt):
        p = evt.GetProperty()
        index = self.pg.GetSelectedPage()
        name = p.GetName()
        newvalue = p.GetValueAsString()

        for n, v in self._value:
            if n == name: v[index] = str(newvalue)
        #evt.Skip()
        self.GetParent().send_event(self, evt)        
        #wx.CallAfter(self.GetTopLevelParent().Show)
        
    def GetValue(self):
        return self._value

    def SetValue(self, value):
        selp = self.pg.GetSelectedPage()        
        self.pg.ClearPage(0)
        self.pg.SelectPage(0)
        self.prop_names = []
        for name, v in value:
            pgp = self.pg.Append(SmootherChoiceProperty(name, 
                                                  choices = SparseSmoother))   
            pgp.SetValue(v[0])
            self.prop_names.append(name)

        self.pg.ClearPage(1)
        self.pg.SelectPage(1)
        for name, v in value:
            pgp = self.pg.Append(SmootherChoiceProperty(name, 
                                                  choices = HypreSmoother))
            pgp.SetValue(v[1])

        self._value = value
        self.pg.SelectPage(selp)
        self.set_column_title()        
'''
from petram.pi.widget_smoother import WidgetSmoother
ll = [[None, None, 99, {"UI":WidgetSmoother, "span":(1,2)}],]
ret= DialogEditList(ll)
print ret
'''



from __future__ import print_function

import numpy as np
import scipy

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('find_dof_map_nd')

#from petram.solver.mumps.hypre_to_mumps import get_HypreParMatrixRow

from petram.mfem_config import use_parallel
if use_parallel:
   from mpi4py import MPI
   num_proc = MPI.COMM_WORLD.size
   myid     = MPI.COMM_WORLD.rank
   from petram.helper.mpi_recipes import *   
   import mfem.par as mfem
else:
   import mfem.ser as mfem
   num_proc = 1
   myid = 0
   def allgather(x): return [x]
   
mapper_debug = False
def find_dof_map_nd(idx1, idx2, transu, transv, transu2, transv2, 
                    fes, engine, tdof, tol=1e-4):
   '''
   find_dof_map

   projection coupling between DoFs. so far, it is tested
   only for triangle Nedelec boundary elemets.

   it maps element nodes of on bdr1 (attribute = idx1) to 
   bdr2(attribute =idx2)

   this version accept two mapping function which mappps
   (x, y, z) to (u, v) plane

   In this version, returned matrix is interpolation constraints P.
   P will be used as follows.
    P^t A P y = P^t f, x = Py
       
   For example..
      if  x2 = x1 and   [A B][x3]  = [x5]
                        [C D][x4]    [x6]
      P should be filled as follows. It starts from identity 
      sparse  matrix and changes the following matrix elements.

      [ 1            ][y1]    [x1]
      [ 1 0          ][y2]    [x2]
      [     1        ][y3]  = [x3] 
      [        1     ][y4]    [x4]
      [     A  B 0   ][y5]    [x5]
      [     C  D   0 ][y6]    [x6]

    row and col of  P^t A P corresponding to x2, x5, x6 will
    be all zero.

    This module actually assemble the transpose of P.
    Diagonal elements should be handled in caller. See em3d_floquet

   '''
   def get_vshape(fes, k1, idx):
       tr1 = fes.GetBdrElementTransformation(k1)
       el = fes.GetBE(k1)
       nodes1 = el.GetNodes()
       m = mfem.DenseMatrix(nodes1.GetNPoints(), tr1.GetSpaceDim())
       tr1.SetIntPoint(nodes1.IntPoint(idx))
       el.CalcVShape(tr1,m)
       return m.GetDataArray()[idx, :]
    
   def get_vshape_all(fes, k1):
       tr1 = fes.GetBdrElementTransformation(k1)
       el = fes.GetBE(k1)
       nodes1 = el.GetNodes()
       m = mfem.DenseMatrix(nodes1.GetNPoints(), tr1.GetSpaceDim())
       shape = []
       for idx in range(nodes1.GetNPoints()):
          tr1.SetIntPoint(nodes1.IntPoint(idx))
          el.CalcVShape(tr1,m)
          shape.append(m.GetDataArray()[idx,:].copy())
       return np.stack(shape)

   def get_all_intpoint2(fes, pt1all, pt2all, pto1all, pto2all, k1all,
                         k2all, sh1all, sh2all, map_1_2,
                         tru, trv, tru2, trv2, 
                         tol, tdof, fesize):
       pt = []
       subvdofs1 = []
       subvdofs2 = []

       map = scipy.sparse.lil_matrix((fesize, fesize), dtype=float)

       num_entry = 0
       num_entry2 = 0       
       num_pts = 0       

       decimals = int(np.abs(np.log10(tol)))
       for k0 in range(len(pt1all)):
           k2 = map_1_2[k0]
           pt1 = pt1all[k0]
           pto1 = pto1all[k0]
           newk1 = k1all[k0]
           sh1 = sh1all[k0]           
           pt2 = pt2all[k2]
           pto2 = pto2all[k2]
           newk2 = k2all[k2]
           sh2 = sh2all[k2]
           for k, p in enumerate(pt1):
               num_pts = num_pts + 1              
               if newk1[k,2] == -1: continue
               if newk1[k,2] in tdof: continue
               if newk1[k,2] in subvdofs1: continue               
               dist = np.sum((pt2-p)**2, 1)
               d = np.where(dist == np.min(dist))[0]
               #dprint1(dist-np.min(dist), len(d))
               if len(d) == 1:
                   '''
                   this factor is not always 1
                   '''
                   num_entry = num_entry + 1                   
                   s = np.sign(newk1[k,1] +0.5)*np.sign(newk2[d,1] + 0.5)
                   d = d[0]
                   
                   p1 = pto1[k]; p2 = pto2[d]
                   delta = np.sum(np.std(p1, 0))/np.sum(np.std(sh1, 0))/10.

                   v1 = np.array([tru(p1) - tru(p1 + delta*sh1[newk1[k, 0]]),
                                  trv(p1) - trv(p1 + delta*sh1[newk1[k, 0]])])
                   v2 = np.array([tru2(p2) - tru2(p2 + delta*sh2[newk2[d, 0]]),
                                  trv2(p2) - trv2(p2 + delta*sh2[newk2[d, 0]])])
                   fac_2 = np.sum(v1*v2)/np.sum(v1*v1)*s 

                   map[newk2[d,2], newk1[k,2]] = np.around(fac_2, decimals)

               elif len(d) == 2:
                   num_entry2 = num_entry2 + 1                  
                   dd = np.argsort(np.sum((pt1 - p)**2, 1))
                   
                   p1 = pto1[dd[0]]; p3 = pto2[d[0]]
                   p2 = pto1[dd[1]]; p4 = pto2[d[1]]
                   delta = np.sum(np.std(p1, 0))/np.sum(np.std(sh1, 0))/10.
                   
                   v1 = np.array([tru(p1) - tru(p1 + delta*sh1[newk1[dd[0], 0]]),
                                  trv(p1) - trv(p1 + delta*sh1[newk1[dd[0], 0]])])
                   v2 = np.array([tru(p2) - tru(p2 + delta*sh1[newk1[dd[1], 0]]),
                                  trv(p2) - trv(p2 + delta*sh1[newk1[dd[1], 0]])])
                   v3 = np.array([tru2(p3) - tru2(p3 + delta*sh2[newk2[d[0], 0]]),
                                  trv2(p3) - trv2(p3 + delta*sh2[newk2[d[0], 0]])])
                   v4 = np.array([tru2(p4) - tru2(p4 + delta*sh2[newk2[d[1], 0]]),
                                  trv2(p4) - trv2(p4 + delta*sh2[newk2[d[1], 0]])])

                   v1 = v1*np.sign(newk1[dd[0], 1] +0.5)
                   v2 = v2*np.sign(newk1[dd[1], 1] +0.5)
                   v3 = v3*np.sign(newk2[d[0], 1] +0.5)
                   v4 = v4*np.sign(newk2[d[1], 1] +0.5)
                   s = np.sign(newk1[k,1] +0.5)*np.sign(newk2[d,1] + 0.5) 

                   def vnorm(v):
                       return v/np.sqrt(np.sum(v**2))
                   v1n = vnorm(v1) ; v2n = vnorm(v2)
                   v3n = vnorm(v3) ; v4n = vnorm(v4)                   

#                   print v1, v2, v3, v4                   
#                   if np.abs((np.abs(np.sum(v1*v3)/np.sum(v1*v1)) - 1) < tol
#                       and np.abs(np.abs(np.sum(v2*v4)/np.sum(v2*v2)) -  1) < tol):
#                   print 'check', np.abs(np.abs(np.sum(v1n*v3n))-1),   np.abs(np.abs(np.sum(v2n*v4n))-1)

                   if (np.abs(np.abs(np.sum(v1n*v3n))-1) < tol and
                       np.abs(np.abs(np.sum(v2n*v4n))-1) < tol):
                       fac1 = np.sum(v1*v3)/np.sum(v1*v1)
                       fac2 = np.sum(v2*v4)/np.sum(v2*v2)
                       #print 'first case', fac1, fac2                  
                       map[newk2[d[0],2], newk1[dd[0],2]] = np.around(fac1, decimals)
                       map[newk2[d[1],2], newk1[dd[1],2]] = np.around(fac2, decimals)
                   elif (np.abs(np.abs(np.sum(v2n*v3n))-1) < tol and
                         np.abs(np.abs(np.sum(v1n*v4n))-1) < tol):
                       fac1 = np.sum(v1*v4)/np.sum(v1*v1)
                       fac2 = np.sum(v2*v3)/np.sum(v2*v2)
                       #print 'second case', fac1, fac2, decimals               
                       map[newk2[d[1],2], newk1[dd[0],2]] = np.around(fac1, decimals)
                       map[newk2[d[0],2], newk1[dd[1],2]] = np.around(fac2, decimals)
                   else:
                       #print 'two shape vector couples!', v1, v2, v3, v4
                       #m2 = np.transpose(np.vstack((v1, v2)))
                       #m1 = np.transpose(np.vstack((v3, v4)))
                       #m = np.dot(np.linalg.inv(m1), m2)
                       m1 = np.transpose(np.vstack((v1, v2)))
                       m2 = np.transpose(np.vstack((v3, v4)))
                       m = np.dot(np.linalg.inv(m1), m2)
                       m = np.around(np.linalg.inv(m), decimals = decimals)
                       #print m
                       #print pt1[dd[0]], pt1[dd[1]], pt2[d[0]], pt2[d[1]]
                       #print v1, v2, v3, v4, newk1[dd[0]][2], newk1[dd[1]][2],newk2[d[0]][2], newk2[d[1]][2]
                       if m[0,0] != 0.: map[newk2[d[0],2], newk1[dd[0],2]] = m[0,0]
                       if m[0,1] != 0.: map[newk2[d[1],2], newk1[dd[0],2]] = m[0,1]
                       if m[1,0] != 0.: map[newk2[d[0],2], newk1[dd[1],2]] = m[1,0]
                       if m[1,1] != 0.: map[newk2[d[1],2], newk1[dd[1],2]] = m[1,1]
               else:
                   raise AssertionError("three dof coupling is not supported. ")
                   pass
           subvdofs1.extend([s for k, v, s in newk1])
           subvdofs2.extend([s for k, v, s in newk2])
       #print len(subvdofs1), len(subvdofs2)
       total_entry2 = sum(allgather(num_entry2))
       total_entry = sum(allgather(num_entry))
       total_pts = sum(allgather(num_pts))
       
       dprint2("map size", map.shape)
       dprint2("local pts/edge entry/mid-face entry", num_pts, " " , num_entry, " ",
               num_entry2)
       dprint1("total pts/edge entry/mid-face entry", total_pts, " " , total_entry, " ",
               total_entry2)
       
       return map


   def find_el_center(fes, ibdr1, tru, trv):
       if len(ibdr1) == 0: return np.empty(shape=(0,2))
       mesh = fes.GetMesh()
       pts = np.vstack([np.mean([(tru(mesh.GetVertexArray(kk)),
                                  trv(mesh.GetVertexArray(kk)))
                                  for kk in mesh.GetBdrElementVertices(k)],0)
                        for k in ibdr1])
       return pts

   def get_vshape_all2(fes, ibdr1):
       if len(ibdr1) == 0:
           tmp = np.stack([get_vshape_all(fes, k) for k in [0]])
           res = np.empty(shape=(0, tmp.shape[1], tmp.shape[2]))
       else:
           res =  np.stack([get_vshape_all(fes, k) for k in ibdr1])
       dprint2('get_vshape_all2', res.shape)
       return res
        
   def get_element_data(k1, debug, tru, trv):
       tr1 = fes.GetBdrElementTransformation(k1)
       nodes1 = fes.GetBE(k1).GetNodes()
       vdof1 = fes.GetBdrElementVDofs(k1)
       pt1 = np.vstack([np.array([tru(tr1.Transform(nodes1.IntPoint(kk))),
                                  trv(tr1.Transform(nodes1.IntPoint(kk)))])
              for kk in range(len(vdof1))])
       pt1o = np.vstack([np.array(tr1.Transform(nodes1.IntPoint(kk)))
                       for kk in range(len(vdof1))])


       #print vdof1
       subvdof1 = [x if x>= 0 else -1-x for x in vdof1]

       if use_parallel:
           subvdof2= [fes.GetLocalTDofNumber(i)
                      for i in subvdof1]


           flag = False
           for k, x in enumerate(subvdof2):
               if x >=0:
                  #subvdof2[k] = get_HypreParMatrixRow(engine.matvec[2][0], x)
                  subvdof2[k] = fes.GetMyTDofOffset()+ x                  
               else:
                  if debug:
                     print('not own')
                     flag = True
           if flag: print(subvdof1, vdof1, subvdof2)

             ## note subdof1 = -1 if it is not owned by the node
       else:
           subvdof2 = subvdof1
       newk1 = np.vstack([(k, xx[0], xx[1])
                for k, xx in enumerate(zip(vdof1, subvdof2))])
       pt1 =  np.vstack([pt1[kk] for kk, v, s in newk1])
       pt1o = np.vstack([pt1o[kk] for kk, v, s in newk1])

       return pt1, newk1, pt1o

   def resolve_nonowned_dof(pt1all, pt2all, k1all, k2all, map_1_2):
       '''
       resolves shadowed DoF
       this is done based on integration point distance.
       It searches a closeest true (non-shadow) DoF point
       '''
       for k in range(len(pt1all)):
          subvdof1 = k1all[k][:,2]
          k2 = map_1_2[k]
          subvdof2 = k2all[k2][:,2]
          pt2 = pt2all[k2]
          check = False
          if -1 in subvdof2:
                check = True
                dprint1('before resolving dof', subvdof2)
          for kk, x in enumerate(subvdof2):
             if x == -1:
                dist = pt2all-pt2[kk]
                dist = np.sqrt(np.sum((dist)**2, -1))
                fdist= dist.flatten()
                isort = np.argsort(fdist)
                dprint3("distances", np.min(dist),fdist[isort[:5]])
                #if fdist[isort[1]] <  np.mean(fdist[isort[2:4]])*1e-6:
                #     minidx = list(isort[:2])
                #else:
                #     minidx = [isort[0]]
                minidx =  np.where(dist.flatten() == np.min(dist.flatten()))[0]
                while all(k2all[:,:,2].flatten()[minidx] == -1):
                    dprint3("distances (non -1 exists?)", fdist[minidx],
                            k2all[:,:,2].flatten()[minidx])
                    minidx = np.hstack((minidx, isort[len(minidx)]))
                dprint3("distances", np.min(dist),fdist[minidx], fdist[isort[:len(minidx)+1]])
                for i in minidx:
                    if k2all[:,:,2].flatten()[i] != -1:
                       subvdof2[kk] = k2all[:,:,2].flatten()[i]
          if check:
              dprint1('resolved dof', k2all[k2][:,2])
              if -1 in subvdof2: raise AssertionError("failed to resolve shadow DoF")
   
   nbe = fes.GetNBE()
   ibdr1 = [i for i in range(nbe) if fes.GetBdrAttribute(i) in idx1]
   ibdr2 = [i for i in range(nbe) if fes.GetBdrAttribute(i) in idx2]
   ct1 = find_el_center(fes, ibdr1, transu, transv)
   ct2 = find_el_center(fes, ibdr2, transu2, transv2)

   arr1 = [get_element_data(k, False, transu, transv) for k in ibdr1]
   arr2 = [get_element_data(k, False, transu2, transv2) for k in ibdr2]
   
   sh1all = get_vshape_all2(fes, ibdr1)
   sh2all = get_vshape_all2(fes, ibdr2)

   # prepare default shape...
   ttt = [get_element_data(k, False, transu, transv) for k in [0]]
   ptzero = np.stack([x for x, y, z in ttt])
   ptozero = np.stack([z for x, y, z in ttt])
   kzero = np.stack([y for x, y, z in ttt])
   
   # pt is on (u, v), pto is (x, y, z)
   pt1all = (np.stack([x for x, y, z in arr1]) if len(arr1)!=0 else
             np.empty(shape=(0,ptzero.shape[1],ptzero.shape[2])))
   pt2all = (np.stack([x for x, y, z in arr2]) if len(arr2)!=0 else
             np.empty(shape=(0,ptzero.shape[1],ptzero.shape[2])))            
   pto1all = (np.stack([z for x, y, z in arr1]) if len(arr1)!=0 else
             np.empty(shape=(0,ptozero.shape[1],ptozero.shape[2])))
   pto2all = (np.stack([z for x, y, z in arr2]) if len(arr2)!=0 else
             np.empty(shape=(0,ptozero.shape[1],ptozero.shape[2])))            

   if use_parallel:
      if MPI.INT.size == 4:
          dtype = np.int32
      else:
          dtype = np.int64
   else:
      dtype = np.int32
   k1all = (np.stack([y for x, y, z in arr1]).astype(dtype) if len(arr1)!=0
            else np.empty(shape=(0,kzero.shape[1],kzero.shape[2])).astype(dtype))
   k2all = (np.stack([y for x, y, z in arr2]).astype(dtype) if len(arr2)!=0
            else np.empty(shape=(0,kzero.shape[1],kzero.shape[2])).astype(dtype))

   if mapper_debug:         
   #   for i in range(num_proc):
      for i in range(1):
          MPI.COMM_WORLD.Barrier()      
          if ( myid == 1): 
             dprint3("checking ptall ", myid)
             for k, x in enumerate(pt1all):
                print(x, k1all[k], sh1all[k])
      for i in range(num_proc):
          MPI.COMM_WORLD.Barrier()      
          if ( myid == i): 
             dprint3("checking ptall2 ", myid)
             for k, x in enumerate(pt2all):
                 print(x, k2all[k], sh2all[k])             

   #print 'k1all', k1all.shape
   #print 'sh1all', sh1all.shape
   fesize = fes.GetNDofs()
   if use_parallel:
       # share ibr2 (destination information among nodes...)
       ct2 =  allgather_vector(ct2, MPI.DOUBLE)
       pt2all =  allgather_vector(pt2all, MPI.DOUBLE)
       pto2all =  allgather_vector(pto2all, MPI.DOUBLE)
       k2all =  allgather_vector(k2all, MPI.INT)
       sh2all =  allgather_vector(sh2all, MPI.DOUBLE)
       fesize = fes.GlobalTrueVSize()
       dprint1(fesize, fes.GetTrueVSize())

   if mapper_debug:         
      MPI.COMM_WORLD.Barrier()                   
      if ( myid == 1 ):
         dprint3("checking pt2all ", myid)
         for k, x in enumerate(pt2all):
              print(x, k2all[k], sh2all[k])
      MPI.COMM_WORLD.Barrier()                   
   #print 'k2all', k2all.shape
   #print 'sh2all', sh2all.shape
   ctr_dist =  np.array([np.min(np.sum((ct2-c)**2, 1)) for c in ct1])
   if ctr_dist.size > 0 and np.max(ctr_dist) > 1e-15:
       print('Center Dist may be too large (check mesh): ' + str(np.max(ctr_dist)))
   map_1_2= [np.argmin(np.sum((ct2-c)**2, 1)) for c in ct1]
   
   if use_parallel:
       resolve_nonowned_dof(pt1all, pt2all, k1all, k2all, map_1_2)

   map =  get_all_intpoint2(fes, pt1all, pt2all, pto1all, pto2all, k1all,
                            k2all, sh1all, sh2all, map_1_2,
                            transu, transv,
                            transu2, transv2,
                            tol, tdof, fesize)

   return map
